"""Construct Parquet database from local file system.

NB: If using a remote Alyx instance it is advisable to generate the cache via the Alyx one_cache
management command, otherwise the resulting cache UUIDs will not match those on the database.

Examples
--------
>>> from one.api import One
>>> cache_dir = 'path/to/data'
>>> make_parquet_db(cache_dir)
>>> one = One(cache_dir=cache_dir)
"""


# -------------------------------------------------------------------------------------------------
# Imports
# -------------------------------------------------------------------------------------------------

import datetime
import uuid
from functools import partial
from pathlib import Path
import warnings
import logging

import pandas as pd
from iblutil.io import parquet
from iblutil.io.hashfile import md5

from one.alf.io import iter_sessions, iter_datasets
from one.alf.files import session_path_parts, get_alf_path
from one.converters import session_record2path

__all__ = ['make_parquet_db', 'remove_missing_datasets']
_logger = logging.getLogger(__name__)

# -------------------------------------------------------------------------------------------------
# Global variables
# -------------------------------------------------------------------------------------------------

SESSIONS_COLUMNS = (
    'id',               # int64
    'lab',
    'subject',
    'date',             # datetime.date
    'number',           # int
    'task_protocol',
    'projects',
)

DATASETS_COLUMNS = (
    'id',               # int64
    'eid',              # int64
    'session_path',     # relative to the root
    'rel_path',         # relative to the session path, includes the filename
    'file_size',        # file size in bytes
    'hash',             # sha1/md5, computed in load function
    'exists',           # bool
)


# -------------------------------------------------------------------------------------------------
# Parsing util functions
# -------------------------------------------------------------------------------------------------

def _ses_str_id(session_path):
    """Returns a str id from a session path in the form '(lab/)subject/date/number'"""
    return Path(*filter(None, session_path_parts(session_path, assert_valid=True))).as_posix()


def _get_session_info(rel_ses_path):
    """Parse a relative session path."""
    out = session_path_parts(rel_ses_path, as_dict=True, assert_valid=True)
    out['id'] = _ses_str_id(rel_ses_path)
    out['date'] = pd.to_datetime(out['date']).date()
    out['number'] = int(out['number'])
    out['task_protocol'] = ''
    out['projects'] = ''
    return out


def _get_dataset_info(full_ses_path, rel_dset_path, ses_eid=None, compute_hash=False):
    rel_ses_path = get_alf_path(full_ses_path)
    full_dset_path = Path(full_ses_path, rel_dset_path).as_posix()
    file_size = Path(full_dset_path).stat().st_size
    ses_eid = ses_eid or _ses_str_id(rel_ses_path)
    return {
        'id': Path(rel_ses_path, rel_dset_path).as_posix(),
        'eid': str(ses_eid),
        'session_path': str(rel_ses_path),
        'rel_path': Path(rel_dset_path).as_posix(),
        'file_size': file_size,
        'hash': md5(full_dset_path) if compute_hash else None,
        'exists': True
    }


def _rel_path_to_uuid(df, id_key='rel_path', base_id=None, keep_old=False):
    base_id = base_id or uuid.uuid1()  # Base hash based on system by default
    toUUID = partial(uuid.uuid3, base_id)  # MD5 hash from base uuid and rel session path string
    if keep_old:
        df[f'{id_key}_'] = df[id_key].copy()
    df[id_key] = df[id_key].apply(lambda x: str(toUUID(x)))
    assert len(df[id_key].unique()) == len(df[id_key])  # WARNING This fails :(
    return df


def _ids_to_uuid(df_ses, df_dsets):
    ns = uuid.uuid1()
    df_dsets = _rel_path_to_uuid(df_dsets, id_key='id', base_id=ns)
    df_ses = _rel_path_to_uuid(df_ses, id_key='id', base_id=ns, keep_old=True)
    # Copy new eids into datasets frame
    df_dsets['eid_'] = df_dsets['eid'].copy()
    df_dsets['eid'] = (df_ses
                       .set_index('id_')
                       .loc[df_dsets['eid'], 'id']
                       .values)
    # Check that the session int IDs in both frames match
    ses_id_set = df_ses.set_index('id_')['id']
    assert (df_dsets
            .set_index('eid_')['eid']
            .drop_duplicates()
            .equals(ses_id_set)), 'session int ID mismatch between frames'

    # Set index
    df_ses = df_ses.set_index('id').drop('id_', axis=1).sort_index()
    df_dsets = df_dsets.set_index(['eid', 'id']).drop('eid_', axis=1).sort_index()

    return df_ses, df_dsets


# -------------------------------------------------------------------------------------------------
# Main functions
# -------------------------------------------------------------------------------------------------

def _metadata(origin):
    """
    Metadata dictionary for Parquet files.

    Parameters
    ----------
    origin : str, pathlib.Path
        Path to full directory, or computer name / db name
    """
    return {
        'date_created': datetime.datetime.now().isoformat(sep=' ', timespec='minutes'),
        'origin': str(origin),
    }


def _make_sessions_df(root_dir) -> pd.DataFrame:
    """
    Given a root directory, recursively finds all sessions and returns a sessions DataFrame

    Parameters
    ----------
    root_dir : str, pathlib.Path
        The folder to look for sessions

    Returns
    -------
    pandas.DataFrame
        A pandas DataFrame of session info
    """
    rows = []
    for full_path in iter_sessions(root_dir):
        # Get the lab/Subjects/subject/date/number part of a file path
        rel_path = get_alf_path(full_path)
        # A dict of session info extracted from path
        ses_info = _get_session_info(rel_path)
        assert set(ses_info.keys()) <= set(SESSIONS_COLUMNS)
        rows.append(ses_info)
    df = pd.DataFrame(rows, columns=SESSIONS_COLUMNS)
    return df


def _make_datasets_df(root_dir, hash_files=False) -> pd.DataFrame:
    """
    Given a root directory, recursively finds all datasets and returns a datasets DataFrame

    Parameters
    ----------
    root_dir : str, pathlib.Path
        The folder to look for sessions
    hash_files : bool
        If True, an MD5 is computed for each file and stored in the 'hash' column

    Returns
    -------
    pandas.DataFrame
        A pandas DataFrame of dataset info
    """
    df = pd.DataFrame([], columns=DATASETS_COLUMNS)
    # Go through sessions and append datasets
    for session_path in iter_sessions(root_dir):
        rows = []
        for rel_dset_path in iter_datasets(session_path):
            file_info = _get_dataset_info(session_path, rel_dset_path, compute_hash=hash_files)
            assert set(file_info.keys()) <= set(DATASETS_COLUMNS)
            rows.append(file_info)
        df = pd.concat((df, pd.DataFrame(rows, columns=DATASETS_COLUMNS)),
                       ignore_index=True, verify_integrity=True)
    return df


def make_parquet_db(root_dir, out_dir=None, hash_ids=True, hash_files=False, lab=None):
    """
    Given a data directory, index the ALF datasets and save the generated cache tables.

    Parameters
    ----------
    root_dir : str, pathlib.Path
        The file directory to index.
    out_dir : str, pathlib.Path
        Optional output directory to save cache tables.  If None, the files are saved into the
        root directory.
    hash_ids : bool
        If True, experiment and dataset IDs will be UUIDs generated from the system and relative
        paths (required for use with ONE API)
    hash_files : bool
        If True, an MD5 hash is computed for each dataset and stored in the datasets table.
        This will substantially increase cache generation time.
    lab : str
        An optional lab name to associate with the data.  If the folder structure
        contains 'lab/Subjects', the lab name will be taken from the folder name.

    Returns
    -------
    pathlib.Path
        The full path of the saved sessions parquet table
    pathlib.Path
        The full path of the saved datasets parquet table
    """
    root_dir = Path(root_dir).resolve()

    # Make the dataframes.
    df_ses = _make_sessions_df(root_dir)
    df_dsets = _make_datasets_df(root_dir, hash_files=hash_files)

    # Add integer id columns
    if hash_ids and len(df_ses) > 0:
        df_ses, df_dsets = _ids_to_uuid(df_ses, df_dsets)

    if lab:  # Fill in lab name field
        assert not df_ses['lab'].any() or (df_ses['lab'] == 'lab').all(), 'lab name conflict'
        df_ses['lab'] = lab

    # Check any files were found
    if df_ses.empty or df_dsets.empty:
        warnings.warn(f'No {"sessions" if df_ses.empty else "datasets"} found', RuntimeWarning)

    # Output directory.
    out_dir = Path(out_dir or root_dir)
    assert out_dir.is_dir()
    assert out_dir.exists()

    # Parquet files to save.
    fn_ses = out_dir / 'sessions.pqt'
    fn_dsets = out_dir / 'datasets.pqt'

    # Parquet metadata.
    metadata = _metadata(root_dir)

    # Save the Parquet files.
    parquet.save(fn_ses, df_ses, metadata)
    parquet.save(fn_dsets, df_dsets, metadata)

    return fn_ses, fn_dsets


def remove_missing_datasets(cache_dir, tables=None, remove_empty_sessions=True, dry=True):
    """
    Remove dataset files and session folders that are not in the provided cache.

    NB: This *does not* remove entries from the cache tables that are missing on disk.
    Non-ALF files are not removed. Empty sessions that exist in the sessions table are not removed.

    Parameters
    ----------
    cache_dir : str, pathlib.Path
    tables : dict[str, pandas.DataFrame], optional
        A dict with keys ('sessions', 'datasets'), containing the cache tables as DataFrames.
    remove_empty_sessions : bool
        Attempt to remove session folders that are empty and not in the sessions table.
    dry : bool
        If true, do not remove anything.

    Returns
    -------
    list
        A sorted list of paths to be removed.
    """
    cache_dir = Path(cache_dir)
    if tables is None:
        tables = {}
        for name in ('datasets', 'sessions'):
            tables[name], _ = parquet.load(cache_dir / f'{name}.pqt')
    to_delete = []
    gen_path = partial(session_record2path, root_dir=cache_dir)
    sessions = sorted(map(lambda x: gen_path(x[1]), tables['sessions'].iterrows()))
    for session_path in iter_sessions(cache_dir):
        rel_session_path = session_path.relative_to(cache_dir).as_posix()
        datasets = tables['datasets'][tables['datasets']['session_path'] == rel_session_path]
        for dataset in iter_datasets(session_path):
            if dataset.as_posix() not in datasets['rel_path']:
                to_delete.append(session_path.joinpath(dataset))
        if session_path not in sessions and remove_empty_sessions:
            to_delete.append(session_path)

    if dry:
        print('The following session and datasets would be removed:', end='\n\t')
        print('\n\t'.join(sorted(map(str, to_delete))))
        return sorted(to_delete)

    # Delete datasets
    for path in to_delete:
        if path.is_file():
            _logger.debug(f'Removing {path}')
            path.unlink()
        else:
            # Recursively remove empty folders
            while path.parent != cache_dir and not next(path.rglob('*'), False):
                _logger.debug(f'Removing {path}')
                path.rmdir()
                path = path.parent

    return sorted(to_delete)



"""Unit tests for the one.registration module."""
import logging
import unittest
import unittest.mock
import string
import random
import datetime
import fnmatch
from io import StringIO
from pathlib import Path

from iblutil.util import Bunch
from requests.exceptions import HTTPError

from one.api import ONE
from one import registration
import one.alf.exceptions as alferr
from one.alf.io import next_num_folder
from . import TEST_DB_1, OFFLINE_ONLY
from one.tests import util


@unittest.skipIf(OFFLINE_ONLY, 'online only test')
class TestRegistrationClient(unittest.TestCase):
    """Test class for RegistrationClient class"""
    one = None
    subject = None
    temp_dir = None

    @classmethod
    def setUpClass(cls) -> None:
        cls.temp_dir = util.set_up_env()
        cls.one = ONE(**TEST_DB_1, cache_dir=cls.temp_dir.name)
        cls.subject = ''.join(random.choices(string.ascii_letters, k=10))
        cls.one.alyx.rest('subjects', 'create', data={'lab': 'mainenlab', 'nickname': cls.subject})
        # Create some files for this subject
        session_path = cls.one.alyx.cache_dir / cls.subject / str(datetime.date.today()) / '001'
        cls.session_path = session_path
        for rel_path in cls.one.list_datasets():
            filepath = session_path.joinpath(rel_path)
            filepath.parent.mkdir(exist_ok=True, parents=True)
            filepath.touch()

    def setUp(self) -> None:
        self.client = registration.RegistrationClient(one=self.one)

    def test_water_administration(self):
        """Test for RegistrationClient.register_water_administration"""
        record = self.client.register_water_administration(self.subject, 35.10000000235)
        self.assertEqual(record['subject'], self.subject)
        self.assertEqual(record['water_administered'], 35.1)
        self.assertEqual(record['water_type'], 'Water')
        self.assertEqual(record['user'], self.one.alyx.user)
        # Create session to associate
        d = {'subject': self.subject,
             'procedures': ['Behavior training/tasks'],
             'type': 'Base',
             'users': [self.one.alyx.user],
             }
        ses = self.one.alyx.rest('sessions', 'create', data=d)
        volume = random.random()
        record = self.client.register_water_administration(self.subject, volume,
                                                           session=ses['url'])
        self.assertEqual(record['subject'], self.subject)
        self.assertEqual(record['session'], ses['url'][-36:])
        # Check validations
        with self.assertRaises(ValueError):
            self.client.register_water_administration(self.subject, volume, session='NaN')
        with self.assertRaises(ValueError):
            self.client.register_water_administration(self.subject, .0)
        with unittest.mock.patch.object(self.client.one, 'to_eid', return_value=None),\
             self.assertRaises(ValueError):
            self.client.register_water_administration(self.subject, 3.6, session=ses['url'])

    def test_register_weight(self):
        """Test for RegistrationClient.register_weight"""
        record = self.client.register_weight(self.subject, 35.10000000235)
        self.assertEqual(record['subject'], self.subject)
        self.assertEqual(record['weight'], 35.1)
        self.assertEqual(record['user'], self.one.alyx.user)
        # Check validations
        with self.assertRaises(ValueError):
            self.client.register_weight(self.subject, 0.0)

    def test_ensure_ISO8601(self):
        """Test for RegistrationClient.ensure_ISO8601"""
        date = datetime.datetime(2021, 7, 14, 15, 53, 15, 525119)
        self.assertEqual(self.client.ensure_ISO8601(date), '2021-07-14T15:53:15.525119')
        self.assertEqual(self.client.ensure_ISO8601(date.date()), '2021-07-14T00:00:00')
        date_str = '2021-07-14T15:53:15.525119'
        self.assertEqual(self.client.ensure_ISO8601(date_str), date_str)
        with self.assertRaises(ValueError):
            self.client.ensure_ISO8601(f'{date:%D}')

    def test_exists(self):
        """Test for RegistrationClient.assert_exists"""
        # Check user endpoint
        with self.assertRaises(alferr.AlyxSubjectNotFound):
            self.client.assert_exists('foobar', 'subjects')
        self.client.assert_exists(self.subject, 'subjects')
        # Check user endpoint with list
        with self.assertRaises(alferr.ALFError) as ex:
            self.client.assert_exists([self.one.alyx.user, 'foobar'], 'users')
        self.assertIn('foobar', str(ex.exception))
        # Check raises non-404
        err = HTTPError()
        err.response = Bunch({'status_code': 500})
        with unittest.mock.patch.object(self.one.alyx, 'get', side_effect=err),\
                self.assertRaises(HTTPError):
            self.client.assert_exists('foobar', 'subjects')

    def test_find_files(self):
        """Test for RegistrationClient.find_files"""
        # Remove a dataset type from the client to check that the dataset(s) are ignored
        existing = (x['filename_pattern'] and any(self.session_path.rglob(x['filename_pattern']))
                    for x in self.client.dtypes)
        removed = self.client.dtypes.pop(next(i for i, x in enumerate(existing) if x))
        files = list(self.client.find_files(self.session_path))
        self.assertEqual(6, len(files))
        self.assertTrue(all(map(Path.is_file, files)))
        # Check removed file pattern not in file list
        self.assertFalse(fnmatch.filter([x.name for x in files], removed['filename_pattern']))

    def test_create_new_session(self):
        """Test for RegistrationClient.create_new_session"""
        # Check register = True
        session_path, eid = self.client.create_new_session(
            self.subject, date='2020-01-01', projects='ibl_neuropixel_brainwide_01'
        )
        expected = self.one.alyx.cache_dir.joinpath(self.subject, '2020-01-01', '001').as_posix()
        self.assertEqual(session_path.as_posix(), expected)
        self.assertIsNotNone(eid)
        # Check projects added to Alyx session
        projects = self.one.get_details(eid, full=True)['projects']
        self.assertEqual(projects, ['ibl_neuropixel_brainwide_01'])
        # Check register = False
        session_path, eid = self.client.create_new_session(
            self.subject, date='2020-01-01', register=False)
        expected = self.one.alyx.cache_dir.joinpath(self.subject, '2020-01-01', '002').as_posix()
        self.assertEqual(session_path.as_posix(), expected)
        self.assertIsNone(eid)

    def test_register_session(self):
        """Test for RegistrationClient.register_session"""
        # Find some datasets to create
        datasets = self.one.list_datasets(self.one.search(dataset='raw')[0])
        session_path = self.one.alyx.cache_dir.joinpath(
            'mainenlab', 'Subjects', self.subject, '2020-01-01', '001'
        )
        # Ensure session exists
        file_list = [session_path.joinpath(x) for x in datasets]
        # Create the files before registering
        for x in file_list:
            x.parent.mkdir(exist_ok=True, parents=True)
            x.touch()

        ses, recs = self.client.register_session(str(session_path),
                                                 end_time='2020-01-02',
                                                 procedures='Behavior training/tasks')
        self.assertTrue(len(ses['data_dataset_session_related']))
        self.assertEqual(len(ses['data_dataset_session_related']), len(recs))
        self.assertTrue(isinstance(ses['procedures'], list) and len(ses['procedures']) == 1)
        self.assertEqual(ses['end_time'], '2020-01-02T00:00:00')

        # Check value error raised when provided lab name doesn't match
        with self.assertRaises(ValueError):
            self.client.register_session(str(session_path), lab='cortexlab')

        # Check updating existing session
        start_time = '2020-01-01T10:36:10'
        ses, _ = self.client.register_session(str(session_path), start_time=start_time)
        self.assertEqual(start_time, ses['start_time'])

    def test_create_sessions(self):
        """Test for RegistrationClient.create_sessions"""
        session_path = self.session_path.parent / next_num_folder(self.session_path.parent)
        session_path.mkdir(parents=True)
        session_path.joinpath('create_me.flag').touch()
        # Should print session path in dry mode
        with unittest.mock.patch('sys.stdout', new_callable=StringIO) as stdout:
            session_paths, ses = self.client.create_sessions(self.one.alyx.cache_dir, dry=True)
            self.assertTrue(str(session_path) in stdout.getvalue())
            self.assertTrue(len(ses) == 1 and ses[0] is None)
            self.assertTrue(session_path.joinpath('create_me.flag').exists())

        # Should find and register session
        session_paths, ses = self.client.create_sessions(self.one.alyx.cache_dir)
        self.assertTrue(len(ses) == 1 and len(session_paths) == 1)
        self.assertFalse(session_path.joinpath('create_me.flag').exists())
        self.assertEqual(ses[0]['number'], int(session_path.parts[-1]))
        self.assertEqual(session_paths[0], session_path)

    def test_register_files(self):
        """Test for RegistrationClient.register_files"""
        # Test a few things not checked in register_session
        session_path, eid = self.client.create_new_session(self.subject)
        # Check registering single file, dry run, default False
        file_name = session_path.joinpath('wheel.position.npy')
        file_name.touch()
        rec = self.client.register_files(str(file_name), default=False, dry=True)
        self.assertIsInstance(rec, dict)
        self.assertFalse(rec['default'])
        self.assertNotIn('id', rec)
        # Add ambiguous dataset type to types list
        self.client.dtypes.append(self.client.dtypes[-1].copy())
        self.client.dtypes[-1]['name'] += '1'
        # Try registering ambiguous / invalid datasets
        ambiguous = self.client.dtypes[-1]['filename_pattern'].replace('*', 'npy')
        files = [session_path.joinpath('wheel.position.xxx'),  # Unknown ext
                 session_path.joinpath('foo.bar.npy'),  # Unknown dtype
                 session_path.joinpath(ambiguous)  # Ambiguous dtype
                 ]
        version = ['1.2.9'] * len(files)
        with self.assertLogs('one.registration', logging.DEBUG) as dbg:
            rec = self.client.register_files(files, versions=version)
            self.assertIn('wheel.position.xxx: No matching extension', dbg.records[0].message)
            self.assertIn('foo.bar.npy: No matching dataset type', dbg.records[1].message)
            self.assertIn(f'{ambiguous}: Multiple matching', dbg.records[2].message)
        self.assertFalse(len(rec))

    def test_instantiation(self):
        """Test RegistrationClient.__init__ with no args"""
        with unittest.mock.patch('one.registration.ONE') as mk:
            client = registration.RegistrationClient()
        self.assertIsInstance(client.one, unittest.mock.MagicMock)
        mk.assert_called_with(cache_rest=None)

    @classmethod
    def tearDownClass(cls) -> None:
        for admin in cls.one.alyx.rest('water-administrations', 'list',
                                       django='subject__nickname,' + cls.subject, no_cache=True):
            cls.one.alyx.rest('water-administrations', 'delete', id=admin['url'][-36:])
        for ses in cls.one.alyx.rest('sessions', 'list', subject=cls.subject, no_cache=True):
            cls.one.alyx.rest('sessions', 'delete', id=ses['url'][-36:])
        cls.one.alyx.rest('subjects', 'delete', id=cls.subject)
        cls.temp_dir.cleanup()


if __name__ == "__main__":
    unittest.main(exit=False)



"""Unit tests for the one.api module

Wherever possible the ONE tests should not rely on an internet connection

Fixture locations:

- The cache tables for the public test instance are in tests/fixtures/
- The test db parameters can be found in tests/fixtures/params/
- Some REST GET requests can be found in tests/fixtures/rest_responses/
- These can be copied over to a temporary directory using the functions in tests/util.py,
  then construct ONE with the directory as cache_dir, mode='local' and silent=True

Imported constants:

- For tests that do require a remote connection use the tests.OFFLINE_ONLY flag in the skipIf
  decorator.
- For testing REST POST requests use TEST_DB_1 (test.alyx.internationalbrainlab.org).
- For testing download functions, use TEST_DB_2 (openalyx.internationalbrainlab.org).

Note ONE and AlyxClient use caching:

- When verifying remote changes via the rest method, use the no_cache flag to ensure the remote
  databaseis queried.  You can clear the cache using AlyxClient.clear_rest_cache(),
  or mock iblutil.io.params.getfile to return a temporary cache directory
- An One object created through the one.api.ONE function, make sure you restore the
  properties to their original state on teardown, or call one.api.ONE.cache_clear()

"""
import datetime
import logging
import time
from pathlib import Path
from itertools import permutations, combinations_with_replacement
from functools import partial
import unittest
from unittest import mock
import tempfile
from uuid import UUID, uuid4
import io

import numpy as np
import pandas as pd
from requests.exceptions import HTTPError
from iblutil.io import parquet
from iblutil.util import Bunch

from one.api import ONE, One, OneAlyx
from one.util import (
    ses2records, validate_date_range, index_last_before, filter_datasets, _collection_spec,
    filter_revision_last_before, parse_id, autocomplete, LazyId, datasets2records
)
import one.params
import one.alf.exceptions as alferr
from . import util
from . import OFFLINE_ONLY, TEST_DB_1, TEST_DB_2  # 1 = TestAlyx; 2 = OpenAlyx


class TestONECache(unittest.TestCase):
    """Test methods that use sessions and datasets tables.
    This class loads the parquet tables from the fixtures and builds a file tree in a temp folder
    """
    tempdir = None

    def setUp(self) -> None:
        self.tempdir = util.set_up_env()
        # Create ONE object with temp cache dir
        self.one = ONE(mode='local', cache_dir=self.tempdir.name)
        # Create dset files from cache
        util.create_file_tree(self.one)

    def tearDown(self) -> None:
        while Path(self.one.cache_dir).joinpath('.cache.lock').exists():
            time.sleep(.1)
        self.tempdir.cleanup()

    def test_list_subjects(self):
        """Test One.list_subejcts"""
        subjects = self.one.list_subjects()
        expected = ['KS005', 'ZFM-01935', 'ZM_1094', 'ZM_1150',
                    'ZM_1743', 'ZM_335', 'clns0730', 'flowers']
        self.assertCountEqual(expected, subjects)

    def test_offline_repr(self):
        """Test for One.offline property"""
        self.assertTrue('offline' in str(self.one))
        self.assertTrue(str(self.tempdir.name) in str(self.one))

    def test_one_search(self):
        """Test for One.search"""
        one = self.one
        # Search subject
        eids = one.search(subject='ZM_335')
        expected = ['3473f9d2-aa5d-41a6-9048-c65d0b7ab97c',
                    'dfe99506-b873-45db-bc93-731f9362e304']
        self.assertEqual(expected, eids)

        # Search lab
        labs = ['mainen', 'cortexlab']
        eids = one.search(laboratory=labs)
        expected = ['d3372b15-f696-4279-9be5-98f15783b5bb',
                    '3473f9d2-aa5d-41a6-9048-c65d0b7ab97c']
        self.assertEqual(len(eids), 25)
        self.assertEqual(expected, eids[:2])

        # Search exact date
        eids = one.search(date='2019-06-07')
        self.assertEqual(eids, ['db524c42-6356-4c61-b236-4967c54d2665'])

        # Search date range
        dates = ['2019-04-01', '2019-04-10']
        eids = one.search(date=dates)
        expected = ['13c99443-01ee-462e-b668-717daa526fc0',
                    'abf5109c-d780-44c8-9561-83e857c7bc01']
        self.assertEqual(len(eids), 9)
        self.assertEqual(expected, eids[:2])

        # Search from a given date
        dates = ['2021-01-01', None]
        eids = one.search(date_range=dates)
        self.assertEqual(eids, ['d3372b15-f696-4279-9be5-98f15783b5bb'])

        # Search datasets
        query = 'spikes.depths'
        eids = one.search(data=query)
        self.assertTrue(eids)
        expected = [
            'd3372b15-f696-4279-9be5-98f15783b5bb',
            'b1c968ad-4874-468d-b2e4-5ffa9b9964e9',
            'cf264653-2deb-44cb-aa84-89b82507028a'
        ]
        self.assertEqual(eids, expected)

        # Filter non-existent
        # Set exist for one of the eids to false
        mask = (one._cache['datasets']['rel_path'].str.contains(query))
        i = one._cache['datasets'][mask].index[0]
        one._cache['datasets'].loc[i, 'exists'] = False

        self.assertTrue(len(eids) == len(one.search(data=query)) + 1)

        # Search task_protocol
        eids = one.search(task='habituation')
        self.assertEqual(eids, ['ac80cd12-49e5-4aff-b5f2-1a718679ceeb'])

        # Search project
        eids = one.search(proj='neuropix')
        self.assertEqual(len(eids), 20)

        # Search number
        number = 1
        eids = one.search(num=number)

        sess_num = self.one._cache.sessions.loc[eids, 'number']
        self.assertTrue(all(sess_num == number))

        number = '002'
        eids = one.search(number=number)

        sess_num = self.one._cache.sessions.loc[eids, 'number']
        self.assertTrue(all(sess_num == int(number)))

        # Empty results
        eids, det = one.search(num=100, subject='KS000', details=True)
        self.assertTrue(len(eids) == 0)
        self.assertIsNone(det)
        # Check works with just one search term
        eids, det = one.search(num=500, details=True)
        self.assertTrue(len(eids) == 0)
        self.assertIsNone(det)

        # Test multiple fields, with short params
        eids = one.search(subj='KS005', date='2019-04-10', num='003', lab='cortexlab')
        self.assertTrue(len(eids) == 1)

        # Test param error validation
        with self.assertRaises(ValueError):
            one.search(dat='2021-03-05')  # ambiguous
        with self.assertRaises(ValueError):
            one.search(user='mister')  # invalid search term

        # Test details parameter
        eids, details = one.search(date='2019-04-10', lab='cortexlab', details=True)
        self.assertEqual(len(eids), len(details))
        self.assertCountEqual(details[0].keys(), self.one._cache.sessions.columns)

        # Test search with integer ids
        util.caches_str2int(one._cache)
        query = 'clusters'
        eids = one.search(data=query)
        self.assertTrue(all(isinstance(x, str) for x in eids))
        self.assertEqual(3, len(eids))

    def test_filter(self):
        """Test one.util.filter_datasets"""
        datasets = self.one._cache.datasets.iloc[:5].copy()
        # Test identity
        verifiable = filter_datasets(datasets, None, None, None,
                                     assert_unique=False, revision_last_before=False)
        self.assertEqual(len(datasets), len(verifiable))

        # Test collection filter
        verifiable = filter_datasets(datasets, None, 'alf', None,
                                     assert_unique=False, revision_last_before=False)
        self.assertEqual(2, len(verifiable))
        with self.assertRaises(alferr.ALFMultipleCollectionsFound):
            filter_datasets(datasets, None, 'raw.*', None, revision_last_before=False)
        # Test filter empty collection
        datasets.rel_path[-1] = '_ibl_trials.rewardVolume.npy'
        verifiable = filter_datasets(datasets, None, '', None, revision_last_before=False)
        self.assertTrue(len(verifiable), 1)

        # Test dataset filter
        verifiable = filter_datasets(datasets, '_iblrig_.*', None, None,
                                     assert_unique=False, revision_last_before=False)
        self.assertEqual(3, len(verifiable))
        with self.assertRaises(alferr.ALFMultipleObjectsFound):
            filter_datasets(datasets, '_iblrig_.*', 'raw_video_data', None,
                            revision_last_before=False)
        # Test list as logical OR
        verifiable = filter_datasets(datasets, ['alf/_ibl_wheel.*', '_ibl_trials.*'], None, None,
                                     assert_unique=False, revision_last_before=False)
        self.assertTrue(all(x.startswith('_ibl_trials') or x.startswith('alf/_ibl_wheel')
                            for x in verifiable.rel_path))

        # Test as dict
        dataset = dict(namespace='ibl', object='trials')
        verifiable = filter_datasets(datasets, dataset, None, None,
                                     assert_unique=False, revision_last_before=False)
        self.assertEqual(1, len(verifiable))
        dataset = dict(timescale='bpod', object='trials')
        verifiable = filter_datasets(self.one._cache.datasets, dataset, None, None,
                                     assert_unique=False, revision_last_before=False)
        self.assertEqual(verifiable['rel_path'].values[0], 'alf/_ibl_trials.intervals_bpod.npy')

        # As dict with list (should act as logical OR)
        dataset = dict(attribute=['time.+', 'raw'])
        verifiable = filter_datasets(datasets, dataset, None, None,
                                     assert_unique=False, revision_last_before=False)
        self.assertEqual(4, len(verifiable))

        # Revisions
        revisions = [
            'alf/probe00/#2020-01-01#/spikes.times.npy',
            'alf/probe00/#2020-08-31#/spikes.times.npy',
            'alf/probe00/spikes.times.npy',
            'alf/probe00/#2021-xx-xx#/spikes.times.npy',
            'alf/probe01/#2020-01-01#/spikes.times.npy'
        ]
        datasets['rel_path'] = revisions

        # Should return last revision before date for each collection/dataset
        revision = '2020-09-06'
        verifiable = filter_datasets(datasets, None, None, revision, assert_unique=False)
        self.assertEqual(2, len(verifiable))
        self.assertTrue(all(x.split('#')[1] < revision for x in verifiable['rel_path']))

        # Should return single dataset with last revision when default specified
        with self.assertRaises(alferr.ALFMultipleRevisionsFound):
            filter_datasets(datasets, '*spikes.times*', 'alf/probe00', None,
                            assert_unique=True, wildcards=True, revision_last_before=True)

        # Should return matching revision
        verifiable = filter_datasets(datasets, None, None, r'2020-08-\d{2}',
                                     assert_unique=False, revision_last_before=False)
        self.assertEqual(1, len(verifiable))
        self.assertTrue(verifiable['rel_path'].str.contains('#2020-08-31#').all())

        # Matches more than one revision; should raise error
        with self.assertRaises(alferr.ALFMultipleRevisionsFound):
            filter_datasets(datasets, None, '.*probe00', r'2020-0[18]-\d{2}',
                            revision_last_before=False)

        # Should return revision that's lexicographically first for each dataset
        verifiable = filter_datasets(datasets, None, None, None, assert_unique=False)
        self.assertEqual(2, len(verifiable))
        actual = tuple(x.split('#')[1] for x in verifiable['rel_path'])
        self.assertEqual(('2021-xx-xx', '2020-01-01'), actual)

        # Should return those without revision
        verifiable = filter_datasets(datasets, None, None, '', assert_unique=False)
        self.assertFalse(verifiable['rel_path'].str.contains('#').any())

        # Should return empty
        verifiable = filter_datasets(datasets, None, '.*01', '', assert_unique=False)
        self.assertEqual(0, len(verifiable))

        verifiable = filter_datasets(datasets, None, '.*01', None, assert_unique=False)
        self.assertEqual(1, len(verifiable))
        self.assertTrue(verifiable['rel_path'].str.contains('#2020-01-01#').all())

        # Should return dataset marked as default
        datasets['default_revision'] = [True] + [False] * 4
        verifiable = filter_datasets(datasets, None, None, None, assert_unique=False)
        self.assertEqual(revisions[0], verifiable.rel_path.values[0])

        # Load dataset with default revision
        verifiable = filter_datasets(datasets, '*spikes.times*', 'alf/probe00', None,
                                     assert_unique=True, wildcards=True, revision_last_before=True)
        self.assertEqual(verifiable.rel_path.to_list(),
                         ['alf/probe00/#2020-01-01#/spikes.times.npy'])
        # When revision_last_before is false, expect multiple objects error
        with self.assertRaises(alferr.ALFMultipleObjectsFound):
            filter_datasets(datasets, '*spikes.times*', 'alf/probe00', None,
                            assert_unique=True, wildcards=True, revision_last_before=False)

    def test_filter_wildcards(self):
        """Test one.util.filter_datasets with wildcards flag set to True"""
        datasets = self.one._cache.datasets.iloc[:5].copy()
        # Test identity
        verifiable = filter_datasets(datasets, '_ibl_*', '*lf', None,
                                     assert_unique=False, wildcards=True)
        self.assertTrue(len(verifiable) == 2)
        # As dict with list (should act as logical OR)
        dataset = dict(attribute=['timestamp?', 'raw'])
        verifiable = filter_datasets(datasets, dataset, None, None,
                                     assert_unique=False, revision_last_before=False,
                                     wildcards=True)
        self.assertEqual(4, len(verifiable))

    def test_list_datasets(self):
        """Test One.list_datasets"""
        # test filename
        dsets = self.one.list_datasets(filename='_ibl_trials*')
        self.assertEqual(len(dsets), 18)
        dsets = self.one.list_datasets(filename='gnagnag')
        self.assertEqual(len(dsets), 0)

        # Test no eid
        dsets = self.one.list_datasets(details=True)
        self.assertEqual(len(dsets), len(self.one._cache.datasets))
        self.assertFalse(dsets is self.one._cache.datasets)

        # Test list for eid
        dsets = self.one.list_datasets('KS005/2019-04-02/001', details=True)
        self.assertEqual(27, len(dsets))

        # Test filters
        filename = {'attribute': ['times', 'intervals'], 'extension': 'npy'}
        dsets = self.one.list_datasets('ZFM-01935/2021-02-05/001', filename)
        self.assertEqual(10, len(dsets))
        self.assertTrue(all(any(y in x for y in ('.times.', '.intervals')) for x in dsets))

        filename['attribute'][0] += '*'  # Include wildcard to match both times and timestamps
        dsets = self.one.list_datasets('ZFM-01935/2021-02-05/001', filename)
        self.assertEqual(13, len(dsets))
        self.assertEqual(3, sum('.timestamps.' in x for x in dsets))

        # Test using str ids as index
        util.caches_str2int(self.one._cache)
        dsets = self.one.list_datasets('KS005/2019-04-02/001')
        self.assertEqual(27, len(dsets))

        # Test empty
        dsets = self.one.list_datasets('FMR019/2021-03-18/002', details=True)
        self.assertIsInstance(dsets, pd.DataFrame)
        self.assertEqual(len(dsets), 0)

        # Test details=False, with and without eid
        for eid in [None, 'KS005/2019-04-02/001']:
            dsets = self.one.list_datasets(eid, details=False)
            self.assertIsInstance(dsets, list)
            self.assertTrue(len(dsets) == np.unique(dsets).size)

    def test_list_collections(self):
        """Test One.list_collections"""
        # Test no eid
        dsets = self.one.list_collections()
        expected = [
            '', 'alf', 'alf/ks2', 'alf/probe00', 'raw_behavior_data', 'raw_ephys_data',
            'raw_ephys_data/probe00', 'raw_passive_data', 'raw_video_data'
        ]
        self.assertCountEqual(expected, dsets)

        # Test details for eid
        dsets = self.one.list_collections('KS005/2019-04-02/001', details=True)
        self.assertIsInstance(dsets, dict)
        self.assertTrue(set(dsets.keys()) <= set(expected))
        self.assertIsInstance(dsets['alf'], pd.DataFrame)
        self.assertTrue(dsets['alf'].rel_path.str.startswith('alf').all())

        # Test empty
        self.assertFalse(len(self.one.list_collections('FMR019/2021-03-18/002', details=True)))
        self.assertFalse(len(self.one.list_collections('FMR019/2021-03-18/002', details=False)))

    def test_list_revisions(self):
        """Test One.list_revisions"""
        # No revisions in cache fixture so generate our own
        revisions_datasets = util.revisions_datasets_table()
        self.one._cache.datasets = pd.concat([self.one._cache.datasets, revisions_datasets])
        eid, _ = revisions_datasets.index[0]

        # Test no eid
        dsets = self.one.list_revisions()
        expected = ['', '2020-01-08', '2021-07-06']
        self.assertCountEqual(expected, dsets)

        # Test details for eid
        dsets = self.one.list_revisions(eid, details=True)
        self.assertIsInstance(dsets, dict)
        self.assertTrue(set(dsets.keys()) <= set(expected))
        self.assertIsInstance(dsets['2020-01-08'], pd.DataFrame)
        self.assertTrue(dsets['2020-01-08'].rel_path.str.contains('#2020-01-08#').all())

        # Test dataset filter
        dsets = self.one.list_revisions(eid, filename='spikes.times.npy', details=True)
        self.assertTrue(dsets['2020-01-08'].rel_path.str.endswith('spikes.times.npy').all())

        # Test collections filter
        dsets = self.one.list_revisions(eid, collection='alf/probe01', details=True)
        self.assertTrue(dsets['2020-01-08'].rel_path.str.startswith('alf/probe01').all())

        # Test revision filter
        revisions = self.one.list_revisions(eid, revision=['202[01]*'])
        self.assertCountEqual(['2020-01-08', '2021-07-06'], revisions)

        # Test empty
        self.assertFalse(len(self.one.list_revisions('FMR019/2021-03-18/002', details=True)))
        self.assertFalse(len(self.one.list_revisions('FMR019/2021-03-18/002', details=False)))

    def test_get_details(self):
        """Test One.get_details"""
        eid = 'aaf101c3-2581-450a-8abd-ddb8f557a5ad'
        det = self.one.get_details(eid)
        self.assertIsInstance(det, pd.Series)
        self.assertEqual('KS005', det.subject)
        self.assertEqual('2019-04-04', str(det.date))
        self.assertEqual(4, det.number)

        # Test details flag
        det = self.one.get_details(eid, full=True)
        self.assertIsInstance(det, pd.DataFrame)
        self.assertTrue('rel_path' in det.columns)

        # Test with int index ids
        util.caches_str2int(self.one._cache)
        det = self.one.get_details(eid)
        self.assertIsInstance(det, pd.Series)

        # Test errors
        with self.assertRaises(alferr.ALFObjectNotFound):
            self.one.get_details(eid.replace('a', 'b'))
        sessions = self.one._cache.sessions
        self.one._cache.sessions = pd.concat([sessions, det.to_frame().T]).sort_index()
        with self.assertRaises(alferr.ALFMultipleObjectsFound):
            self.one.get_details(eid)

    def test_index_type(self):
        """Test One._index_type"""
        self.assertIs(str, self.one._index_type())
        util.caches_str2int(self.one._cache)
        self.assertIs(int, self.one._index_type())
        self.one._cache.datasets.reset_index(inplace=True)
        with self.assertRaises(IndexError):
            self.one._index_type('datasets')

    def test_load_dataset(self):
        """Test One.load_dataset"""
        eid = 'KS005/2019-04-02/001'
        # Check download only
        file = self.one.load_dataset(eid, '_ibl_wheel.position.npy', download_only=True)
        self.assertIsInstance(file, Path)

        # Check loading data
        np.save(str(file), np.arange(3))  # Make sure we have something to load
        dset = self.one.load_dataset(eid, '_ibl_wheel.position.npy')
        self.assertTrue(np.all(dset == np.arange(3)))

        # Check collection filter
        file = self.one.load_dataset(eid, '_iblrig_leftCamera.timestamps.ssv',
                                     download_only=True, collection='raw_video_data')
        self.assertIsNotNone(file)

        # Test errors
        # ID not in cache
        fake_id = self.one.to_eid(eid).replace('b', 'a')
        with self.assertRaises(alferr.ALFObjectNotFound):
            self.one.load_dataset(fake_id, '_iblrig_leftCamera.timestamps.ssv')
        file.unlink()
        with self.assertRaises(alferr.ALFObjectNotFound):
            self.one.load_dataset(eid, '_iblrig_leftCamera.timestamps.ssv')

        # Check loading without extension
        file = self.one.load_dataset(eid, '_ibl_wheel.position', download_only=True)
        self.assertTrue(str(file).endswith('wheel.position.npy'))

    def test_load_datasets(self):
        """Test One.load_datasets"""
        eid = 'KS005/2019-04-02/001'
        # Check download only
        dsets = ['_ibl_wheel.position.npy', '_ibl_wheel.timestamps.npy']
        files, meta = self.one.load_datasets(eid, dsets, download_only=True, assert_present=False)
        self.assertIsInstance(files, list)
        self.assertTrue(all(isinstance(x, Path) for x in files))

        # Check loading data and missing dataset
        dsets = ['_ibl_wheel.position.npy', '_ibl_wheel.timestamps_bpod.npy']
        np.save(str(files[0]), np.arange(3))  # Make sure we have something to load
        data, meta = self.one.load_datasets(eid, dsets, download_only=False, assert_present=False)
        self.assertEqual(2, len(data))
        self.assertEqual(2, len(meta))
        self.assertTrue(np.all(data[0] == np.arange(3)))

        # Check assert_present raises error
        with self.assertRaises(alferr.ALFObjectNotFound):
            self.one.load_datasets(eid, dsets, assert_present=True)

        # Check collection and revision filters
        dsets = ['_ibl_wheel.position.npy', '_ibl_wheel.timestamps.npy']
        files, meta = self.one.load_datasets(eid, dsets, collections='alf', revisions=[None, None],
                                             download_only=True, assert_present=False)
        self.assertTrue(all(files))

        files, meta = self.one.load_datasets(eid, dsets, collections=['alf', ''],
                                             download_only=True, assert_present=False)
        self.assertIsNone(files[-1])

        # Check validations
        with self.assertRaises(ValueError):
            self.one.load_datasets(eid, dsets, collections=['alf', '', 'foo'])
        with self.assertRaises(TypeError):
            self.one.load_datasets(eid, 'spikes.times')
        with self.assertRaises(alferr.ALFObjectNotFound):
            self.one.load_datasets('ff812ca5-ce60-44ac-b07e-66c2c37e98eb', dsets)
        with self.assertLogs(logging.getLogger('one.api'), 'WARNING'):
            data, meta = self.one.load_datasets('ff812ca5-ce60-44ac-b07e-66c2c37e98eb', dsets,
                                                assert_present=False)
        self.assertIsNone(data)
        self.assertEqual(0, len(meta))
        self.assertIsNone(self.one.load_datasets(eid, [])[0])
        with self.assertRaises(alferr.ALFObjectNotFound):
            self.one.load_datasets(eid, dsets, collections='none', assert_present=True)

        # Check behaviour when data are not downloaded for any reason
        with mock.patch.object(self.one, '_check_filesystem',
                               side_effect=lambda x, **kwargs: [None] * len(x)):
            with self.assertLogs(logging.getLogger('one.api'), 'WARNING'):
                self.one.load_datasets(eid, dsets, collections='alf', assert_present=False)
            with self.assertRaises(alferr.ALFObjectNotFound):
                self.one.load_datasets(eid, dsets, collections='alf', assert_present=True)

        # Check loading without extensions
        # Check download only
        dsets = ['_ibl_wheel.position.npy', '_ibl_wheel.timestamps']
        files, meta = self.one.load_datasets(eid, dsets, download_only=True)
        self.assertTrue(all(isinstance(x, Path) for x in files))

    def test_load_dataset_from_id(self):
        """Test One.load_dataset_from_id"""
        id = np.array([[-9204203870374650458, -6411285612086772563]])
        file = self.one.load_dataset_from_id(id, download_only=True)
        self.assertIsInstance(file, Path)
        expected = 'ZFM-01935/2021-02-05/001/alf/probe00/_phy_spikes_subset.waveforms.npy'
        self.assertTrue(file.as_posix().endswith(expected))

        # Details
        _, details = self.one.load_dataset_from_id(id, download_only=True, details=True)
        self.assertIsInstance(details, pd.Series)

        # Load file content with str id
        s_id, = parquet.np2str(id)
        data = np.arange(3)
        np.save(str(file), data)  # Ensure data to load
        dset = self.one.load_dataset_from_id(s_id)
        self.assertTrue(np.array_equal(dset, data))

        # Load file content with UUID
        dset = self.one.load_dataset_from_id(UUID(s_id))
        self.assertTrue(np.array_equal(dset, data))

        # Load with int ids as index
        util.caches_str2int(self.one._cache)
        dset = self.one.load_dataset_from_id(s_id)
        self.assertTrue(np.array_equal(dset, data))

        # Test errors
        # ID not in cache
        with self.assertRaises(alferr.ALFObjectNotFound):
            self.one.load_dataset_from_id(s_id.replace('a', 'b'))
        # File missing
        file.unlink()
        with self.assertRaises(alferr.ALFObjectNotFound):
            self.one.load_dataset_from_id(s_id)
        # Duplicate ids in cache
        details.name = (8737712210713458643, -4920882604093676133, *id.flatten().tolist())
        datasets = self.one._cache.datasets
        self.one._cache.datasets = pd.concat([datasets, details.to_frame().T]).sort_index()
        with self.assertRaises(alferr.ALFMultipleObjectsFound):
            self.one.load_dataset_from_id(s_id)

    def test_load_object(self):
        """Test One.load_object"""
        eid = 'aaf101c3-2581-450a-8abd-ddb8f557a5ad'
        files = self.one.load_object(eid, 'wheel', download_only=True)
        self.assertEqual(len(files), 3)
        self.assertTrue(all(isinstance(x, Path) for x in files))

        # Save some data into the files
        N = 10  # length of data
        for f in files:
            np.save(str(f), np.random.rand(N))
        wheel = self.one.load_object(eid, 'wheel')
        self.assertIsInstance(wheel, dict)
        self.assertCountEqual(wheel.keys(), ('position', 'velocity', 'timestamps'))
        self.assertTrue(
            all(x.size == N for x in wheel.values())
        )

        # Test errors
        with self.assertRaises(alferr.ALFObjectNotFound):
            self.one.load_object(eid, 'spikes')
        # Test behaviour with missing session
        with self.assertRaises(alferr.ALFObjectNotFound):
            self.one.load_object(eid.replace('a', 'b'), 'wheel')
        # Test missing files on disk
        [f.unlink() for f in files]
        with self.assertRaises(alferr.ALFObjectNotFound):
            self.one.load_object(eid, 'wheel')
        # Check the three wheel datasets set to exist = False in cache
        self.assertFalse(self.one._cache.datasets.loc[(eid,), 'exists'].all())

        eid = 'ZFM-01935/2021-02-05/001'
        with self.assertRaises(alferr.ALFMultipleCollectionsFound):
            self.one.load_object(eid, 'ephysData_g0_t0')
        with self.assertRaises(alferr.ALFMultipleObjectsFound):
            self.one.load_object(eid, '*Camera')

    def test_load_collection(self):
        """Test One.load_collection"""
        # Check download_only output
        eid = 'aaf101c3-2581-450a-8abd-ddb8f557a5ad'
        files = self.one.load_collection(eid, 'alf', download_only=True)
        self.assertEqual(len(files), 18)
        self.assertTrue(all(isinstance(x, Path) for x in files))
        # Check load
        alf = self.one.load_collection(eid, 'alf', attribute='*time*')
        self.assertCountEqual(alf.keys(), ('trials', 'wheel'))
        self.assertIsInstance(alf, Bunch)
        self.assertIn('feedback_times', alf.trials)
        # Check object filter
        alf = self.one.load_collection(eid, 'alf', object='trials')
        self.assertNotIn('wheel', alf)
        # Check errors
        with self.assertRaises(alferr.ALFError):
            self.one.load_collection(eid, '')
        with self.assertRaises(alferr.ALFObjectNotFound):
            self.one.load_collection(eid, 'alf', object='foo')
        # Should raise error when no objects found on disk
        with mock.patch.object(self.one, '_check_filesystem',
                               side_effect=lambda x, **kwargs: [None] * len(x)),\
                self.assertRaises(alferr.ALFObjectNotFound) as ex:
            self.one.load_collection(eid, 'alf')
            self.assertIn('not found on disk', str(ex))

    def test_load_cache(self):
        """Test One._load_cache"""
        # Test loading unsorted table with no id index set
        df = self.one._cache['datasets'].reset_index()
        info = self.one._cache['_meta']['raw']['datasets']
        with tempfile.TemporaryDirectory() as tdir:
            # Loading from empty dir
            self.one.load_cache(tdir)
            self.assertTrue(self.one._cache['_meta']['expired'])
            # Save unindexed
            parquet.save(Path(tdir) / 'datasets.pqt', df, info)
            del self.one._cache['datasets']
            self.one.load_cache(tdir)
            self.assertIsInstance(self.one._cache['datasets'].index, pd.MultiIndex)
            # Save shuffled
            id_keys = ['eid', 'id']
            df[id_keys] = np.random.permutation(df[id_keys])
            assert not df.set_index(id_keys).index.is_monotonic_increasing
            parquet.save(Path(tdir) / 'datasets.pqt', df, info)
            del self.one._cache['datasets']
            self.one.load_cache(tdir)
            self.assertTrue(self.one._cache['datasets'].index.is_monotonic_increasing)
            # Save a parasitic table that will not be loaded
            pd.DataFrame().to_parquet(Path(tdir).joinpath('gnagna.pqt'))
            with self.assertLogs(logging.getLogger('one.api'), logging.WARNING) as log:
                self.one.load_cache(tdir)
                self.assertTrue('gnagna.pqt' in log.output[0])
            # Save table with missing id columns
            df.drop(id_keys, axis=1, inplace=True)
            parquet.save(Path(tdir) / 'datasets.pqt', df, info)
            with self.assertRaises(KeyError):
                self.one.load_cache(tdir)

    def test_refresh_cache(self):
        """Test One.refresh_cache"""
        self.one._cache.datasets = self.one._cache.datasets.iloc[0:0].copy()
        prev_loaded = self.one._cache['_meta']['loaded_time']
        for mode in ('auto', 'local', 'remote'):
            with self.subTest("Refresh modes", mode=mode):
                loaded = self.one.refresh_cache(mode)
                self.assertFalse(len(self.one._cache.datasets))
                self.assertEqual(prev_loaded, loaded)
        loaded = self.one.refresh_cache('refresh')
        self.assertTrue(len(self.one._cache.datasets))
        self.assertTrue(loaded > prev_loaded)
        self.one.cache_expiry = datetime.timedelta()  # Immediately expire
        self.one._cache.datasets = self.one._cache.datasets.iloc[0:0].copy()
        self.one.refresh_cache('auto')
        self.assertTrue(len(self.one._cache.datasets))
        with self.assertRaises(ValueError):
            self.one.refresh_cache('double')

    def test_save_cache(self):
        """Test one.util.save_cache"""
        self.one._cache['_meta'].pop('modified_time', None)
        # Should be no cache save as it's not been modified
        with tempfile.TemporaryDirectory() as tdir:
            self.one._save_cache(save_dir=tdir)
            self.assertFalse(any(Path(tdir).glob('*')))
            # Should save two tables
            self.one._cache['_meta']['modified_time'] = datetime.datetime.now()
            self.one._save_cache(save_dir=tdir)
            self.assertEqual(2, len(list(Path(tdir).glob('*.pqt'))))
            # Load with One and check modified time is preserved
            raw_modified = One(cache_dir=tdir)._cache['_meta']['raw']['datasets']['date_modified']
            expected = self.one._cache['_meta']['modified_time'].strftime('%Y-%m-%d %H:%M')
            self.assertEqual(raw_modified, expected)

    def test_update_cache_from_records(self):
        """Test One._update_cache_from_records"""
        # Update with single record (pandas.Series), one exists, one doesn't
        session = self.one._cache.sessions.iloc[0].squeeze()
        session.name = str(uuid4())  # New record
        dataset = self.one._cache.datasets.iloc[0].squeeze()
        dataset['exists'] = not dataset['exists']
        self.one._update_cache_from_records(sessions=session, datasets=dataset)
        self.assertTrue(session.name in self.one._cache.sessions.index)
        updated, = dataset['exists'] == self.one._cache.datasets.loc[dataset.name, 'exists']
        self.assertTrue(updated)

        # Update a number of records
        datasets = self.one._cache.datasets.iloc[:3].copy()
        datasets.loc[:, 'exists'] = ~datasets.loc[:, 'exists']
        # Make one of the datasets a new record
        idx = datasets.index.values
        idx[-1] = (idx[-1][0], str(uuid4()))
        datasets.index = pd.MultiIndex.from_tuples(idx)
        self.one._update_cache_from_records(datasets=datasets)
        self.assertTrue(idx[-1] in self.one._cache.datasets.index)
        verifiable = self.one._cache.datasets.loc[datasets.index.values, 'exists']
        self.assertTrue(np.all(verifiable == datasets.loc[:, 'exists']))

        # Check behaviour when columns don't match
        datasets.loc[:, 'exists'] = ~datasets.loc[:, 'exists']
        datasets['extra_column'] = True
        self.one._cache.datasets['new_column'] = False
        self.addCleanup(self.one._cache.datasets.drop, 'new_column', axis=1, inplace=True)
        with self.assertRaises(AssertionError):
            self.one._update_cache_from_records(datasets=datasets, strict=True)
        self.one._update_cache_from_records(datasets=datasets)
        verifiable = self.one._cache.datasets.loc[datasets.index.values, 'exists']
        self.assertTrue(np.all(verifiable == datasets.loc[:, 'exists']))

        # Check fringe cases
        with self.assertRaises(KeyError):
            self.one._update_cache_from_records(unknown=datasets)
        self.assertIsNone(self.one._update_cache_from_records(datasets=None))

    def test_save_loaded_ids(self):
        """Test One.save_loaded_ids and logic within One._check_filesystem"""
        self.one.record_loaded = True  # Turn on saving UUIDs
        self.one._cache.pop('_loaded_datasets', None)  # Ensure we start with a clean slate
        eid = 'd3372b15-f696-4279-9be5-98f15783b5bb'
        files = self.one.load_object(eid, 'trials', download_only=True)
        # Check datasets added to list
        self.assertIn('_loaded_datasets', self.one._cache)
        loaded = self.one._cache['_loaded_datasets']
        self.assertEqual(len(files), len(loaded))
        # Ensure all are from the same session
        eids = self.one._cache.datasets.loc[(slice(None), loaded), ].index.get_level_values(0)
        self.assertTrue(np.all(eids == eid))

        # Test loading a dataset that doesn't exist
        dset = self.one.list_datasets(eid, filename='*trials*', details=True).iloc[-1]
        dset['rel_path'] = dset['rel_path'].replace('.npy', '.pqt')
        dset.name = (eid, str(uuid4()))
        old_cache = self.one._cache['datasets']
        try:
            self.one._cache['datasets'] = self.one._cache.datasets.append(dset)
            dsets = [dset['rel_path'], '_ibl_trials.feedback_times.npy']
            new_files, rec = self.one.load_datasets(eid, dsets, assert_present=False)
            loaded = self.one._cache['_loaded_datasets']
            # One dataset is already in the list (test for duplicates) and other doesn't exist
            self.assertEqual(len(files), len(loaded), 'No new UUIDs should have been added')
            self.assertIn(rec[1]['id'], loaded)  # Already in list
            self.assertEqual(len(loaded), len(np.unique(loaded)))
            self.assertNotIn(dset.name[1], loaded)  # Wasn't loaded as doesn't exist on disk
        finally:
            self.one._cache['datasets'] = old_cache

        # Test saving the loaded datasets list
        ids, filename = self.one.save_loaded_ids(clear_list=False)
        self.assertTrue(loaded is ids)
        self.assertTrue(filename.exists())
        # Load from file
        ids = pd.read_csv(filename)
        self.assertCountEqual(ids['dataset_uuid'], loaded)
        self.assertTrue(self.one._cache['_loaded_datasets'].size, 'List unexpectedly cleared')

        # Test as session UUIDs
        ids, filename = self.one.save_loaded_ids(sessions_only=True, clear_list=False)
        self.assertCountEqual([eid], ids)
        self.assertEqual(pd.read_csv(filename)['session_uuid'][0], eid)

        # Test int IDs.
        self.one._cache['_loaded_datasets'] = parquet.str2np(self.one._cache['_loaded_datasets'])
        # IDs should be cast to string
        with self.assertRaises(NotImplementedError):
            self.one.save_loaded_ids(clear_list=False, sessions_only=True)

        # Check dataset int IDs and clear list
        ids, _ = self.one.save_loaded_ids(clear_list=True)
        self.assertCountEqual(loaded, ids)
        self.assertFalse(self.one._cache['_loaded_datasets'].size, 'List not cleared')

        # Test clear list and warn on empty
        with self.assertWarns(Warning):
            ids, filename = self.one.save_loaded_ids()
        self.assertEqual(ids, [])
        self.assertIsNone(filename)


@unittest.skipIf(OFFLINE_ONLY, 'online only test')
class TestOneAlyx(unittest.TestCase):
    """
    This could be an offline test.  Would need to add /docs REST cache fixture.
    """
    tempdir = None
    one = None

    @classmethod
    def setUpClass(cls) -> None:
        cls.tempdir = util.set_up_env()
        with mock.patch('one.params.iopar.getfile', new=partial(util.get_file, cls.tempdir.name)):
            # util.setup_test_params(token=True)
            cls.one = OneAlyx(
                **TEST_DB_1,
                cache_dir=cls.tempdir.name,
                mode='local'
            )

    def tearDown(self) -> None:
        self.one.mode = 'local'

    def test_type2datasets(self):
        """Test OneAlyx.type2datasets"""
        self.one.mode = 'remote'
        eid = 'cf264653-2deb-44cb-aa84-89b82507028a'
        # when the dataset is at the root, there shouldn't be the separator
        dsets = self.one.type2datasets(eid, 'eye.blink')
        self.assertCountEqual(dsets, ['eye.blink.npy'])
        # test multiples datasets with collections
        eid = '8dd0fcb0-1151-4c97-ae35-2e2421695ad7'
        dtypes = ['trials.feedback_times', '_iblrig_codeFiles.raw']
        dsets = self.one.type2datasets(eid, dtypes)
        expected = ['alf/_ibl_trials.feedback_times.npy',
                    'raw_behavior_data/_iblrig_codeFiles.raw.zip']
        self.assertCountEqual(dsets, expected)
        # this returns a DataFrame
        dsets = self.one.type2datasets(eid, dtypes, details=True)
        self.assertIsInstance(dsets, pd.DataFrame)
        # check validation
        with self.assertRaises(TypeError):
            self.one.type2datasets(eid, 14)

    def test_dataset2type(self):
        """Test for OneAlyx.dataset2type"""
        # Test dataset ID
        did = np.array([[-1058666951852871669, -6012002505064768322]])

        # Check assertion in local mode
        with self.assertRaises(AssertionError):
            self.one.dataset2type(did)

        self.one.mode = 'remote'
        dset_type = self.one.dataset2type(did)
        self.assertEqual('wheelMoves.peakAmplitude', dset_type)
        # Check with tuple int
        dset_type = self.one.dataset2type(tuple(did.squeeze().tolist()))
        self.assertEqual('wheelMoves.peakAmplitude', dset_type)
        # Check with str id
        did, = parquet.np2str(did)
        dset_type = self.one.dataset2type(did)
        self.assertEqual('wheelMoves.peakAmplitude', dset_type)

        dset_type = self.one.dataset2type('_ibl_wheelMoves.peakAmplitude.npy')
        self.assertEqual('wheelMoves.peakAmplitude', dset_type)

        bad_id = np.array([[-1058666951852871669, -6012002505064768312]])
        with self.assertRaises(ValueError):
            self.one.dataset2type(bad_id)

    def test_ses2records(self):
        """Test one.util.ses2records"""
        eid = '8dd0fcb0-1151-4c97-ae35-2e2421695ad7'
        ses = self.one.alyx.rest('sessions', 'read', id=eid)
        session, datasets = ses2records(ses)

        # Verify returned tables are compatible with cache tables
        self.assertIsInstance(session, pd.Series)
        self.assertIsInstance(datasets, pd.DataFrame)
        self.assertEqual(session.name, eid)
        self.assertCountEqual(session.keys(), self.one._cache['sessions'].columns)
        self.assertEqual(len(datasets), len(ses['data_dataset_session_related']))
        expected = [x for x in self.one._cache['datasets'].columns] + ['default_revision']
        self.assertCountEqual(expected, datasets.columns)
        self.assertEqual(tuple(datasets.index.names), ('eid', 'id'))
        self.assertTrue(datasets.default_revision.all())

        # Check int_id as True
        session, datasets = ses2records(ses, int_id=True)
        self.assertEqual(session.name, (-7544566139326771059, -2928913016589240914))
        self.assertEqual(tuple(datasets.index.names), ('eid_0', 'eid_1', 'id_0', 'id_1'))

        # Check behaviour when no datasets present
        ses['data_dataset_session_related'] = []
        _, datasets = ses2records(ses)
        self.assertIsNone(datasets)

    def test_datasets2records(self):
        """Test one.util.datasets2records"""
        eid = '8dd0fcb0-1151-4c97-ae35-2e2421695ad7'
        dsets = self.one.alyx.rest('datasets', 'list', session=eid)
        datasets = datasets2records(dsets)

        # Verify returned tables are compatible with cache tables
        self.assertIsInstance(datasets, pd.DataFrame)
        self.assertTrue(len(datasets) >= len(dsets))
        expected = self.one._cache['datasets'].columns
        self.assertCountEqual(expected, (x for x in datasets.columns if x != 'default_revision'))
        self.assertEqual(tuple(datasets.index.names), ('eid', 'id'))

        # Check behaviour when ind_id is True
        datasets = datasets2records(dsets, int_id=True)
        self.assertEqual(tuple(datasets.index.names), ('eid_0', 'eid_1', 'id_0', 'id_1'))

        # Test single input
        dataset = datasets2records(dsets[0])
        self.assertTrue(len(dataset) == 1)
        # Test records when data missing
        dsets[0]['file_records'][0]['exists'] = False
        empty = datasets2records(dsets[0])
        self.assertTrue(isinstance(empty, pd.DataFrame) and len(empty) == 0)

    def test_pid2eid(self):
        """Test OneAlyx.pid2eid"""
        pid = 'b529f2d8-cdae-4d59-aba2-cbd1b5572e36'
        eid, collection = self.one.pid2eid(pid, query_type='remote')
        self.assertEqual('fc737f3c-2a57-4165-9763-905413e7e341', eid)
        self.assertEqual('probe00', collection)
        with self.assertRaises(NotImplementedError):
            self.one.pid2eid(pid, query_type='local')

    @unittest.mock.patch('sys.stdout', new_callable=io.StringIO)
    def test_describe_revision(self, mock_stdout):
        """Test OneAlyx.describe_revision"""
        self.one.mode = 'remote'
        record = {
            'name': str(datetime.date.today()) + 'a',
            'description': 'Lorem ipsum dolor sit amet, consectetur adipiscing elit.'
        }
        try:
            self.one.alyx.rest('revisions', 'read', id=record['name'], no_cache=True)
        except HTTPError:
            self.one.alyx.rest('revisions', 'create', data=record)
        self.one.describe_revision(record['name'])
        self.assertEqual(record['description'], mock_stdout.getvalue().strip())
        self.one.describe_revision('foobar')
        self.assertTrue('not found' in mock_stdout.getvalue())

        # Check full kwarg
        full = self.one.describe_revision(record['name'], full=True)
        self.assertIsInstance(full, dict)

        # Check raises non-404 error
        err = HTTPError()
        err.response = Bunch({'status_code': 500})
        with mock.patch.object(self.one.alyx, 'get', side_effect=err),\
                self.assertRaises(HTTPError):
            self.one.describe_revision(record['name'])

    @unittest.mock.patch('sys.stdout', new_callable=io.StringIO)
    def test_describe_dataset(self, mock_stdout):
        """Test OneAlyx.describe_dataset.
        NB This could be offline: REST responses in fixtures.
        """
        self.one.mode = 'remote'
        # Test all datasets
        dset_types = self.one.describe_dataset()
        self.assertEqual(7, len(dset_types))
        self.assertEqual('unknown', dset_types[0]['name'])

        # Test dataset type
        out = self.one.describe_dataset('wheel.velocity')
        expected = 'Signed velocity of wheel'
        self.assertTrue(expected in mock_stdout.getvalue())
        self.assertEqual(expected, out['description'])

        # Test dataset name
        expected = 'amplitude of the wheel move'
        out = self.one.describe_dataset('_ibl_wheelMoves.peakAmplitude.npy')
        self.assertTrue(expected in mock_stdout.getvalue())
        self.assertEqual(expected, out['description'])

        # Test unknown dataset name
        with self.assertRaises(ValueError):
            self.one.describe_dataset('_ibl_foo.bar.baz')

    def test_url_from_path(self):
        """Test OneAlyx.path2url"""
        file = Path(self.tempdir.name).joinpath('cortexlab', 'Subjects', 'KS005', '2019-04-04',
                                                '004', 'alf', '_ibl_wheel.position.npy')
        url = self.one.path2url(file)
        self.assertTrue(url.startswith(self.one.alyx._par.HTTP_DATA_SERVER))
        self.assertTrue('91546fc6-b67c-4a69-badc-5e66088519c4' in url)
        # Check remote mode
        url = self.one.path2url(file, query_type='remote')
        # Local data server path ends with '/public'; remote path does not
        self.assertTrue(url.startswith(self.one.alyx._par.HTTP_DATA_SERVER.rsplit('/', 1)[0]))

        file = file.parent / '_fake_obj.attr.npy'
        self.assertIsNone(self.one.path2url(file))
        # Check remote mode  FIXME Different behaviour between remote and local modes
        with self.assertRaises(alferr.ALFObjectNotFound):
            self.one.path2url(file, query_type='remote')

    def test_url_from_record(self):
        """Test ConversionMixin.record2url"""
        idx = (slice(None), '91546fc6-b67c-4a69-badc-5e66088519c4')
        dataset = self.one._cache['datasets'].loc[idx, :]
        url = self.one.record2url(dataset.squeeze())
        expected = ('https://ibl.flatironinstitute.org/public/'
                    'cortexlab/Subjects/KS005/2019-04-04/004/alf/'
                    '_ibl_wheel.position.91546fc6-b67c-4a69-badc-5e66088519c4.npy')
        self.assertEqual(expected, url)

    def test_load_cache(self):
        """Test loading the remote cache"""
        self.one.alyx.silent = False  # For checking log
        self.one._cache._meta['expired'] = True
        try:
            with self.assertLogs(logging.getLogger('one.api'), logging.INFO) as lg:
                with mock.patch.object(self.one.alyx, 'get', side_effect=HTTPError()):
                    self.one.load_cache(clobber=True)
                self.assertEqual('remote', self.one.mode)
                self.assertRegex(lg.output[0], 'cache over .+ old')
                self.assertTrue('Failed to load' in lg.output[1])

                with mock.patch.object(self.one.alyx, 'get', side_effect=ConnectionError()):
                    self.one.load_cache(clobber=True)
                    self.assertEqual('local', self.one.mode)
                self.assertTrue('Failed to connect' in lg.output[-1])

            cache_info = {'min_api_version': '200.0.0'}
            # Check version verification
            with mock.patch.object(self.one.alyx, 'get', return_value=cache_info),\
                    self.assertWarns(UserWarning):
                self.one.load_cache(clobber=True)

        finally:  # Restore properties
            self.one.mode = 'auto'
            self.one.alyx.silent = True

    def test_check_filesystem(self):
        """Test for One._check_filesystem.
        Most is already covered by other tests, this just checks that it can deal with dataset
        dicts as input.
        """
        eid = 'd3372b15-f696-4279-9be5-98f15783b5bb'
        dataset_type = 'probes.description'
        dsets = self.one.alyx.rest('datasets', 'list', session=eid, dataset_type=dataset_type)
        # Create file on disk
        file_ = self.one.eid2path(eid).joinpath('alf', 'probes.description.json')
        file_.parent.mkdir(parents=True)
        file_.touch()
        # Test method
        file, = self.one._check_filesystem(dsets)
        self.assertIsNotNone(file)

    @mock.patch('boto3.Session')
    def test_download_aws(self, boto3_mock):
        """ Tests for the OneAlyx._download_aws method"""
        N = 5
        dsets = self.one._cache['datasets'].iloc[:N].copy()
        dsets['exists_aws'] = True

        self.one.mode = 'auto'  # Can't download in local mode

        # Return a file size so progress bar callback hook functions
        file_object = mock.MagicMock()
        file_object.content_length = 1024
        boto3_mock().resource().Object.return_value = file_object

        # Mock _download_dataset for safety: method should not be called
        with mock.patch.object(self.one, '_download_dataset') as fallback_method:
            out_paths = self.one._download_datasets(dsets)
            fallback_method.assert_not_called()
        self.assertEqual(len(out_paths), N, 'Returned list should equal length of input datasets')
        self.assertTrue(all(isinstance(x, Path) for x in out_paths))
        # These values come from REST cache fixture
        boto3_mock.assert_called_with(aws_access_key_id='ABCDEF', aws_secret_access_key='shhh',
                                      region_name=None)
        ((bucket, path), _), *_ = boto3_mock().resource().Object.call_args_list
        self.assertEqual(bucket, 's3_bucket')
        self.assertTrue(dsets['rel_path'][0].split('.')[0] in path)
        self.assertTrue(dsets.index[0][1] in path, 'Dataset UUID not in filepath')

        # Should fall back to usual method if any datasets do not exist on AWS
        dsets['exists_aws'] = False
        with mock.patch.object(self.one, '_download_dataset') as fallback_method:
            self.one._download_datasets(dsets)
            fallback_method.assert_called()

    @classmethod
    def tearDownClass(cls) -> None:
        cls.tempdir.cleanup()


@unittest.skipIf(OFFLINE_ONLY, 'online only test')
class TestOneRemote(unittest.TestCase):
    """Test remote queries"""
    def setUp(self) -> None:
        self.one = OneAlyx(**TEST_DB_2)
        self.eid = '4ecb5d24-f5cc-402c-be28-9d0f7cb14b3a'

    def test_online_repr(self):
        """Tests OneAlyx.__repr__"""
        self.assertTrue('online' in str(self.one))
        self.assertTrue(TEST_DB_2['base_url'] in str(self.one))

    def test_list_datasets(self):
        """Test OneAlyx.list_datasets"""
        # Test list for eid
        # Ensure remote by making local datasets table empty
        self.addCleanup(self.one.load_cache)
        self.one._cache['datasets'] = self.one._cache['datasets'].iloc[0:0].copy()

        dsets = self.one.list_datasets(self.eid, details=True, query_type='remote')
        self.assertEqual(108, len(dsets))

        # Test missing eid
        dsets = self.one.list_datasets('FMR019/2021-03-18/002', details=True, query_type='remote')
        self.assertIsInstance(dsets, pd.DataFrame)
        self.assertEqual(len(dsets), 0)

        # Test empty datasets
        with mock.patch('one.util.ses2records', return_value=(None, None)):
            dsets = self.one.list_datasets(self.eid, details=True, query_type='remote')
            self.assertIsInstance(dsets, pd.DataFrame)
            self.assertEqual(len(dsets), 0)

        # Test details=False, with eid
        dsets = self.one.list_datasets(self.eid, details=False, query_type='remote')
        self.assertIsInstance(dsets, list)
        self.assertEqual(108, len(dsets))

        # Test with other filters
        dsets = self.one.list_datasets(self.eid, collection='*probe*', filename='*channels*',
                                       details=False, query_type='remote')
        self.assertEqual(13, len(dsets))
        self.assertTrue(all(x in y for x in ('probe', 'channels') for y in dsets))

        with self.assertWarns(Warning):
            self.one.list_datasets(query_type='remote')

    def test_search(self):
        """Test OneAlyx.search"""
        eids = self.one.search(subject='SWC_043', query_type='remote')
        self.assertIn(self.eid, list(eids))

        eids, det = self.one.search(subject='SWC_043', query_type='remote', details=True)
        correct = len(det) == len(eids) and 'url' in det[0] and det[0]['url'].endswith(eids[0])
        self.assertTrue(correct)

        # Check minimum set of keys present (these are present in One.search output)
        expected = {'lab', 'subject', 'date', 'number', 'projects'}
        self.assertTrue(det[0].keys() >= expected)
        # Test dataset search with Django
        eids = self.one.search(subject='SWC_043', dataset=['probes.description'], number=1,
                               django='data_dataset_session_related__collection__iexact,alf',
                               query_type='remote')
        self.assertCountEqual(eids, [self.eid])

        # Test date range
        eids = self.one.search(subject='SWC_043', date='2020-09-21', query_type='remote')
        self.assertCountEqual(eids, [self.eid])

        eids = self.one.search(date=[datetime.date(2020, 9, 21), datetime.date(2020, 9, 22)],
                               lab='hoferlab', query_type='remote')
        self.assertCountEqual(eids, [self.eid])

        # Test limit arg and LazyId
        eids = self.one.search(date='2020-03-23', limit=2, query_type='remote')
        self.assertIsInstance(eids, LazyId)
        self.assertTrue(all(len(x) == 36 for x in eids))

        # Test laboratory kwarg
        eids = self.one.search(laboratory='hoferlab', query_type='remote')
        self.assertIn(self.eid, list(eids))

        eids = self.one.search(lab='hoferlab', query_type='remote')
        self.assertIn(self.eid, list(eids))

    def test_load_dataset(self):
        """Test OneAlyx.load_dataset"""
        file = self.one.load_dataset(self.eid, '_spikeglx_sync.channels.npy',
                                     collection='raw_ephys_data', query_type='remote',
                                     download_only=True)
        self.assertIsInstance(file, Path)
        self.assertTrue(file.as_posix().endswith('raw_ephys_data/_spikeglx_sync.channels.npy'))
        # Test validations
        with self.assertRaises(alferr.ALFMultipleCollectionsFound):
            self.one.load_dataset(self.eid, 'spikes.clusters', query_type='remote')
        with self.assertRaises(alferr.ALFMultipleObjectsFound):
            self.one.load_dataset(self.eid, '_iblrig_*Camera.raw', query_type='remote')
        with self.assertRaises(alferr.ALFObjectNotFound):
            self.one.load_dataset(self.eid, '_iblrig_encoderEvents.raw.ssv',
                                  collection='alf', query_type='remote')

    def test_load_object(self):
        """Test OneAlyx.load_object"""
        files = self.one.load_object(self.eid, 'wheel',
                                     collection='alf', query_type='remote',
                                     download_only=True)
        self.assertIsInstance(files[0], Path)
        self.assertTrue(
            files[0].as_posix().endswith('SWC_043/2020-09-21/001/alf/_ibl_wheel.position.npy')
        )

    def test_get_details(self):
        """Test OneAlyx.get_details"""
        det = self.one.get_details(self.eid, query_type='remote')
        self.assertIsInstance(det, dict)
        self.assertEqual('SWC_043', det['subject'])
        self.assertEqual('ibl_neuropixel_brainwide_01', det['projects'])
        self.assertEqual('2020-09-21', str(det['date']))
        self.assertEqual(1, det['number'])
        self.assertNotIn('data_dataset_session_related', det)

        # Test list
        det = self.one.get_details([self.eid, self.eid], full=True)
        self.assertIsInstance(det, list)
        self.assertIn('data_dataset_session_related', det[0])


@unittest.skipIf(OFFLINE_ONLY, 'online only test')
class TestOneDownload(unittest.TestCase):
    """Test downloading datasets using OpenAlyx"""
    tempdir = None
    one = None

    def setUp(self) -> None:
        self.tempdir = tempfile.TemporaryDirectory()
        self.patch = mock.patch('one.params.iopar.getfile',
                                new=partial(util.get_file, self.tempdir.name))
        self.patch.start()
        self.one = OneAlyx(**TEST_DB_2, cache_dir=self.tempdir.name)
        self.fid = '17ab5b57-aaf6-4016-9251-66daadc200c7'  # File record of channels.brainLocation
        self.eid = 'aad23144-0e52-4eac-80c5-c4ee2decb198'

    def test_download_datasets(self):
        """Test OneAlyx._download_dataset, _download_file and _dset2url"""
        det = self.one.get_details(self.eid, True)
        rec = next(x for x in det['data_dataset_session_related']
                   if 'channels.brainLocation' in x['dataset_type'])
        # FIXME hack because data_url may be AWS
        from one.alf.files import get_alf_path
        rec['data_url'] = self.one.alyx.rel_path2url(get_alf_path(rec['data_url']))
        # FIXME order may not be stable, this only works
        file = self.one._download_dataset(rec)
        self.assertIsInstance(file, Path)
        self.assertTrue(file.exists())

        url = rec['data_url']
        file = self.one._download_dataset(url)
        self.assertIsNotNone(file)

        rec = self.one.alyx.get(rec['url'])
        file = self.one._download_dataset(rec)
        self.assertIsNotNone(file)

        # Check behaviour when hash mismatch
        self.one.alyx.silent = False  # So we can check for warning
        file_hash = rec['hash'].replace('a', 'd')
        # Check three things:
        # 1. The mismatch should be logged at the debug level
        # 2. As we don't have permission to update this db we should see a failure warning
        with self.assertLogs(logging.getLogger('one.api'), logging.DEBUG), \
                self.assertWarns(Warning, msg=f'files/{self.fid}'):
            self.one._download_dataset(rec, hash=file_hash)
        self.one.alyx.silent = True  # Remove console clutter

        # Check JSON field added
        # 3. The files endpoint should be called with a 'mismatch_hash' json key
        fr = [{'url': f'files/{self.fid}', 'json': None}]
        with mock.patch.object(self.one.alyx, '_generic_request', return_value=fr) as patched:
            self.one._download_dataset(rec, hash=file_hash)
            args, kwargs = patched.call_args
            self.assertEqual(kwargs.get('data', {}), {'json': {'mismatch_hash': True}})

        # Check keep_uuid kwarg
        # FIXME Another hack: for this to work the file records order must be correct.
        fi = next(i for i, x in enumerate(rec['file_records'])
                  if x['data_url'].startswith(self.one.alyx._par.HTTP_DATA_SERVER))
        if fi != 0:
            rec['file_records'] = [rec['file_records'].pop(fi), *rec['file_records']]
        file = self.one._download_dataset(rec, keep_uuid=True)
        self.assertEqual(str(file).split('.')[2], rec['url'].split('/')[-1])

        # Check list input
        files = self.one._download_dataset([rec] * 2)
        self.assertIsInstance(files, list)
        self.assertTrue(all(isinstance(x, Path) for x in files))

        # Check Series input
        r_ = datasets2records(rec, int_id=True).squeeze()
        file = self.one._download_dataset(r_)
        self.assertIn('channels.brainLocation', file.as_posix())

        # Check behaviour when URL invalid
        did = rec['url'].split('/')[-1]
        self.assertTrue(self.one._cache.datasets.loc[(slice(None), did), 'exists'].all())
        for fr in rec['file_records']:
            fr['data_url'] = None
        file = self.one._download_dataset(rec)
        self.assertIsNone(file)
        self.assertFalse(self.one._cache.datasets.loc[(slice(None), did), 'exists'].all())
        # With multiple dsets
        files = self.one._download_dataset([rec, rec])
        self.assertTrue(all(x is None for x in files))
        self.one._cache.datasets.loc[(slice(None), did), 'exists'] = True  # Reset values

        # Check with invalid path
        path = self.one.cache_dir.joinpath('lab', 'Subjects', 'subj', '2020-01-01', '001',
                                           'spikes.times.npy')
        with self.assertLogs(logging.getLogger('one.api'), logging.WARNING):
            file = self.one._download_dataset(path)
            self.assertIsNone(file)

        rec = self.one.list_datasets(self.eid, details=True)
        rec = rec[rec.rel_path.str.contains('pykilosort/channels.brainLocation')]
        rec['exists_aws'] = False  # Ensure we use FlatIron for this
        files = self.one._download_datasets(rec)
        self.assertFalse(None in files)

        # Check update cache when id is int and cache table ids are str
        int_id, = parquet.str2np(np.array(rec.index.tolist())).tolist()
        self.one._download_dataset({'data_url': None, 'id': np.array(int_id)})
        exists, = self.one._cache.datasets.loc[(slice(None), rec.index[0]), 'exists']
        self.assertFalse(exists, 'failed to update dataset cache with str index')
        self.one._cache.datasets.loc[(slice(None), rec.index[0]), 'exists'] = True  # Reset values

        # Check works with int index
        util.caches_str2int(self.one._cache)
        self.assertIsNotNone(self.one._download_dataset(files[0]))
        self.assertIsNotNone(self.one._download_datasets(rec))
        # and when dataset missing
        with mock.patch.object(self.one, 'record2url', return_value=None):
            self.assertIsNone(self.one._download_dataset(rec.squeeze()))

        exists, = self.one._cache.datasets.loc[(slice(None), slice(None), *int_id), 'exists']
        self.assertFalse(exists, 'failed to update dataset cache with str index')

    def test_tag_mismatched_file_record(self):
        """Test for OneAlyx._tag_mismatched_file_record.
        This method is also tested in test_download_datasets.
        """
        did = '4a1500c2-60f3-418f-afa2-c752bb1890f0'
        url = f'https://example.com/channels.brainLocationIds_ccf_2017.{did}.npy'
        data = [{'json': {'mismatch_hash': False}, 'url': f'https://example.com/files/{did}'}]
        with mock.patch.object(self.one.alyx, 'rest', return_value=data) as mk:
            self.one._tag_mismatched_file_record(url)
        data[0]['json']['mismatch_hash'] = True
        mk.assert_called_with('files', 'partial_update', id=did, data={'json': data[0]['json']})

    def tearDown(self) -> None:
        try:
            # In case we did indeed have remote REST permissions, try resetting the json field
            self.one.alyx.rest('files', 'partial_update', id=self.fid, data={'json': None})
        except HTTPError as ex:
            if ex.errno != 403:
                raise ex
        self.patch.stop()
        self.tempdir.cleanup()


class TestOneSetup(unittest.TestCase):
    """Test parameter setup upon ONE instantiation and calling setup methods"""
    def setUp(self) -> None:
        self.tempdir = tempfile.TemporaryDirectory()
        self.addCleanup(self.tempdir.cleanup)
        self.get_file = partial(util.get_file, self.tempdir.name)
        # Change default cache dir to temporary directory
        patch = mock.patch('one.params.CACHE_DIR_DEFAULT', Path(self.tempdir.name))
        patch.start()
        self.addCleanup(patch.stop)

    def test_local_cache_setup_prompt(self):
        """Test One.setup"""
        path = Path(self.tempdir.name).joinpath('subject', '2020-01-01', '1', 'spikes.times.npy')
        path.parent.mkdir(parents=True)
        path.touch()
        with mock.patch('builtins.input', return_value=self.tempdir.name):
            one_obj = One.setup()
        self.assertCountEqual(one_obj.list_datasets(), ['spikes.times.npy'])

        # Check prompts warns about cache existing
        path.parent.joinpath('spikes.clusters.npy').touch()
        one_obj = One.setup(cache_dir=self.tempdir.name, silent=True)
        self.assertEqual(1, len(one_obj.list_datasets()))
        with mock.patch('builtins.input', side_effect=['n', 'y']):
            one_obj = One.setup(cache_dir=self.tempdir.name)  # Reply no to cache overwrite
            self.assertEqual(1, len(one_obj.list_datasets()))
            one_obj = One.setup(cache_dir=self.tempdir.name)  # Reply yes to cache overwrite
            self.assertEqual(2, len(one_obj.list_datasets()))

        # Test with no prompt
        one_obj = One.setup(cache_dir=self.tempdir.name, silent=True)
        assert len(one_obj.list_datasets()) == 2

    def test_setup_silent(self):
        """Test setting up parameters with silent flag.
        - Mock getfile to return temp dir as param file location
        - Mock input function as fail safe in case function erroneously prompts user for input
        """
        with mock.patch('iblutil.io.params.getfile', new=self.get_file),\
                mock.patch('one.params.input', new=self.assertFalse):
            one_obj = ONE(silent=True, mode='local', password=TEST_DB_2['password'])
            self.assertEqual(one_obj.alyx.base_url, one.params.default().ALYX_URL)

        # Check param files were saved
        self.assertEqual(len(list(Path(self.tempdir.name).rglob('.caches'))), 1)
        client_pars = Path(self.tempdir.name).rglob(f'.{one_obj.alyx.base_url.split("/")[-1]}')
        self.assertEqual(len(list(client_pars)), 1)

        # Check uses defaults on second instantiation
        with mock.patch('iblutil.io.params.getfile', new=self.get_file):
            one_obj = ONE(mode='local')
            self.assertEqual(one_obj.alyx.base_url, one.params.default().ALYX_URL)

        # Check saves base_url arg
        with self.subTest('Test setup with base URL'):
            if OFFLINE_ONLY:
                self.skipTest('Requires remote db connection')
            with mock.patch('iblutil.io.params.getfile', new=self.get_file):
                one_obj = ONE(**TEST_DB_1)
                self.assertEqual(one_obj.alyx.base_url, TEST_DB_1['base_url'])
                params_url = one.params.get(client=TEST_DB_1['base_url']).ALYX_URL
                self.assertEqual(params_url, one_obj.alyx.base_url)

    def test_setup_username(self):
        """Test setting up parameters with a provided username.
        - Mock getfile to return temp dir as param file location
        - Mock input function as fail safe in case function erroneously prompts user for input
        - Mock requests.post returns a fake user authentication response
        """
        credentials = {'username': 'foobar', 'password': '123'}
        with mock.patch('iblutil.io.params.getfile', new=self.get_file),\
                mock.patch('one.params.input', new=self.assertFalse),\
                mock.patch('one.webclient.requests.post') as req_mock:
            req_mock().json.return_value = {'token': 'shhh'}
            # In remote mode the cache endpoint will not be queried
            one_obj = ONE(base_url='https://test.alyx.internationalbrainlab.org',
                          silent=True, mode='remote', **credentials)
            params_username = one.params.get(client=TEST_DB_1['base_url']).ALYX_LOGIN
            self.assertEqual(params_username, one_obj.alyx.user)
            self.assertEqual(credentials['username'], one_obj.alyx.user)
            _, kwargs = req_mock.call_args
            self.assertEqual(kwargs.get('data', {}), credentials)

            # Reinstantiate as a different user
            one_obj = ONE(base_url='https://test.alyx.internationalbrainlab.org',
                          username='baz', password='123', mode='remote')
            self.assertEqual(one_obj.alyx.user, 'baz')
            self.assertEqual(one_obj.alyx.user, one_obj.alyx._par.ALYX_LOGIN)
            # After initial set up the username in the pars file should remain unchanged
            params_username = one.params.get(client=TEST_DB_1['base_url']).ALYX_LOGIN
            self.assertNotEqual(one_obj.alyx.user, params_username)

    @unittest.skipIf(OFFLINE_ONLY, 'online only test')
    def test_static_setup(self):
        """Test OneAlyx.setup"""
        with mock.patch('iblutil.io.params.getfile', new=self.get_file),\
                mock.patch('one.webclient.getpass', return_value='international'):
            one_obj = OneAlyx.setup(silent=True)
        self.assertEqual(one_obj.alyx.base_url, one.params.default().ALYX_URL)

    def test_setup(self):
        """Test one.params.setup"""
        url = TEST_DB_1['base_url']

        def mock_input(prompt):
            if prompt.lower().startswith('warning'):
                if not getattr(mock_input, 'conflict_warn', False):    # Checks both responses
                    mock_input.conflict_warn = True
                    return 'y'
                return 'n'
            elif 'download cache' in prompt.lower():
                return Path(self.tempdir.name).joinpath('downloads').as_posix()
            elif 'url' in prompt.lower():
                return url
            else:
                return 'mock_input'
        one.params.input = mock_input
        one.params.getpass = lambda prompt: 'mock_pwd'
        one.params.print = lambda text: 'mock_print'
        # Mock getfile function to return a path to non-existent file instead of usual one pars
        with mock.patch('iblutil.io.params.getfile', new=self.get_file):
            one_obj = OneAlyx(mode='local',
                              username=TEST_DB_1['username'],
                              password=TEST_DB_1['password'])
            pars = one.params.get(url)
            self.assertFalse('ALYX_PWD' in pars.as_dict())
        self.assertEqual(one_obj.alyx._par.ALYX_URL, url)
        client_pars = Path(self.tempdir.name).rglob(f'.{one_obj.alyx.base_url.split("/")[-1]}')
        self.assertEqual(len(list(client_pars)), 1)
        # Save ALYX_PWD into params and see if setup modifies it
        with mock.patch('iblutil.io.params.getfile', new=self.get_file):
            one.params.save(pars.set('ALYX_PWD', 'foobar'), url)
            one.params.setup(url)
            self.assertEqual(one.params.get(url).ALYX_PWD, 'mock_pwd')

        # Check conflict warning
        with mock.patch('iblutil.io.params.getfile', new=self.get_file):
            OneAlyx(mode='local',
                    base_url=TEST_DB_2['base_url'],
                    username=TEST_DB_2['username'],
                    password=TEST_DB_2['password'])
        self.assertTrue(getattr(mock_input, 'conflict_warn', False))

    def test_patch_params(self):
        """Test patching legacy params to the new location"""
        # Save some old-style params
        old_pars = one.params.default().set('HTTP_DATA_SERVER', 'openalyx.org')
        # Save a REST query in the old location
        old_rest = Path(self.tempdir.name, '.one', '.rest', old_pars.ALYX_URL[8:], 'https')
        old_rest.mkdir(parents=True, exist_ok=True)
        old_rest.joinpath('1baff95c2d0e31059720a3716ad5b5a34b61a207').touch()

        with mock.patch('iblutil.io.params.getfile', new=self.get_file):
            one.params.setup(silent=True)
            one.params.save(old_pars, old_pars.ALYX_URL)
            one_obj = ONE(base_url=old_pars.ALYX_URL, mode='local')
        self.assertEqual(one_obj.alyx._par.HTTP_DATA_SERVER, one.params.default().HTTP_DATA_SERVER)
        self.assertFalse(Path(self.tempdir.name, '.rest').exists())
        self.assertTrue(any(one_obj.alyx.cache_dir.joinpath('.rest').glob('*')))

    def test_one_factory(self):
        """Tests the ONE class factory"""
        with mock.patch('iblutil.io.params.getfile', new=self.get_file),\
                mock.patch('one.params.input', new=self.assertFalse):
            # Cache dir not in client cache map; use One (light)
            one_obj = ONE(cache_dir=self.tempdir.name)
            self.assertIsInstance(one_obj, One)

            # The offline param was given, raise deprecation warning (via log)
            # with self.assertLogs(logging.getLogger('ibllib'), logging.WARNING):
            #     ONE(offline=True, cache_dir=self.tempdir.name)
            with self.assertWarns(DeprecationWarning):
                ONE(offline=True, cache_dir=self.tempdir.name)

            with self.subTest('ONE setup with database URL'):
                if OFFLINE_ONLY:
                    self.skipTest('Requires remote db connection')
                # No cache dir provided; use OneAlyx (silent setup mode)
                one_obj = ONE(silent=True, mode='local', password=TEST_DB_2['password'])
                self.assertIsInstance(one_obj, OneAlyx)

                # The cache dir is in client cache map; use OneAlyx
                one_obj = ONE(cache_dir=one_obj.alyx.cache_dir, mode='local')
                self.assertIsInstance(one_obj, OneAlyx)

                # A db URL was provided; use OneAlyx
                # mode = 'local' ensures we don't download cache (could also set cache_dir)
                one_obj = ONE(**TEST_DB_1, mode='local')
                self.assertIsInstance(one_obj, OneAlyx)


class TestOneMisc(unittest.TestCase):
    """Test functions in one.util"""
    def test_validate_date_range(self):
        """Test one.util.validate_date_range"""
        # Single string date
        actual = validate_date_range('2020-01-01')  # On this day
        expected = (pd.Timestamp('2020-01-01 00:00:00'),
                    pd.Timestamp('2020-01-01 23:59:59.999000'))
        self.assertEqual(actual, expected)

        # Single datetime.date object
        actual = validate_date_range(pd.Timestamp('2020-01-01 00:00:00').date())
        self.assertEqual(actual, expected)

        # Single pandas Timestamp
        actual = validate_date_range(pd.Timestamp(2020, 1, 1))
        self.assertEqual(actual, expected)

        # Array of two datetime64
        actual = validate_date_range(np.array(['2022-01-30', '2022-01-30'],
                                              dtype='datetime64[D]'))
        expected = (pd.Timestamp('2022-01-30 00:00:00'), pd.Timestamp('2022-01-30 00:00:00'))
        self.assertEqual(actual, expected)

        # From date (lower bound)
        actual = validate_date_range(['2020-01-01'])  # from date
        self.assertEqual(actual[0], pd.Timestamp('2020-01-01 00:00:00'))
        dt = actual[1] - pd.Timestamp.now()
        self.assertTrue(dt.days > 10 * 365)

        actual = validate_date_range(['2020-01-01', None])  # from date
        self.assertEqual(actual[0], pd.Timestamp('2020-01-01 00:00:00'))
        dt = actual[1] - pd.Timestamp.now()
        self.assertTrue(dt.days > 10 * 365)  # Upper bound at least 60 years in the future

        # To date (upper bound)
        actual = validate_date_range([None, '2020-01-01'])  # up to date
        self.assertEqual(actual[1], pd.Timestamp('2020-01-01 00:00:00'))
        dt = pd.Timestamp.now().date().year - actual[0].date().year
        self.assertTrue(dt > 60)  # Lower bound at least 60 years in the past

        self.assertIsNone(validate_date_range(None))
        with self.assertRaises(ValueError):
            validate_date_range(['2020-01-01', '2019-09-06', '2021-10-04'])

    def test_index_last_before(self):
        """Test one.util.index_last_before"""
        revisions = ['2021-01-01', '2020-08-01', '', '2020-09-30']
        verifiable = index_last_before(revisions, '2021-01-01')
        self.assertEqual(0, verifiable)

        verifiable = index_last_before(revisions, '2020-09-15')
        self.assertEqual(1, verifiable)

        verifiable = index_last_before(revisions, '')
        self.assertEqual(2, verifiable)

        self.assertIsNone(index_last_before([], '2009-01-01'))

        verifiable = index_last_before(revisions, None)
        self.assertEqual(0, verifiable, 'should return most recent')

    def test_collection_spec(self):
        """Test one.util._collection_spec"""
        # Test every combination of input
        inputs = []
        _collection = {None: '({collection}/)?', '': '', '-': '{collection}/'}
        _revision = {None: '(#{revision}#/)?', '': '', '-': '#{revision}#/'}
        combs = combinations_with_replacement((None, '', '-'), 2)
        [inputs.extend(set(permutations(x))) for x in combs]
        for collection, revision in inputs:
            with self.subTest(collection=collection, revision=revision):
                verifiable = _collection_spec(collection, revision)
                expected = _collection[collection] + _revision[revision]
                self.assertEqual(expected, verifiable)

    def test_revision_last_before(self):
        """Test one.util.filter_revision_last_before"""
        datasets = util.revisions_datasets_table()
        df = datasets[datasets.rel_path.str.startswith('alf/probe00')].copy()
        verifiable = filter_revision_last_before(df,
                                                 revision='2020-09-01', assert_unique=False)
        self.assertTrue(len(verifiable) == 2)

        # Test assert unique
        with self.assertRaises(alferr.ALFMultipleRevisionsFound):
            filter_revision_last_before(df, revision='2020-09-01', assert_unique=True)

        # Test with default revisions
        df['default_revision'] = False
        with self.assertLogs(logging.getLogger('one.util')):
            verifiable = filter_revision_last_before(df.copy(), assert_unique=False)
        self.assertTrue(len(verifiable) == 2)

        # Should have fallen back on lexicographical ordering
        self.assertTrue(verifiable.rel_path.str.contains('#2021-07-06#').all())
        with self.assertRaises(alferr.ALFError):
            filter_revision_last_before(df.copy(), assert_unique=True)

        # Add unique default revisions
        df.iloc[[0, 4], -1] = True
        verifiable = filter_revision_last_before(df.copy(), assert_unique=True)
        self.assertTrue(len(verifiable) == 2)
        self.assertCountEqual(verifiable['rel_path'], df['rel_path'].iloc[[0, 4]])

        # Add multiple default revisions
        df['default_revision'] = True
        with self.assertRaises(alferr.ALFMultipleRevisionsFound):
            filter_revision_last_before(df.copy(), assert_unique=True)

    def test_parse_id(self):
        """Test one.util.parse_id"""
        obj = unittest.mock.MagicMock()  # Mock object to decorate
        obj.to_eid.return_value = 'parsed_id'  # Method to be called
        input = 'subj/date/num'  # Input id to pass to `to_eid`
        parse_id(obj.method)(obj, input)
        obj.to_eid.assert_called_with(input)
        obj.method.assert_called_with(obj, 'parsed_id')

        # Test raises value error when None returned
        obj.to_eid.return_value = None  # Simulate failure to parse id
        with self.assertRaises(ValueError):
            parse_id(obj.method)(obj, input)

    def test_autocomplete(self):
        """Test one.util.autocomplete"""
        search_terms = ('subject', 'date_range', 'dataset', 'dataset_type')
        self.assertEqual('subject', autocomplete('Subj', search_terms))
        self.assertEqual('dataset', autocomplete('dataset', search_terms))
        with self.assertRaises(ValueError):
            autocomplete('dtypes', search_terms)
        with self.assertRaises(ValueError):
            autocomplete('dat', search_terms)

    def test_LazyID(self):
        """Test one.util.LazyID"""
        uuids = [
            'c1a2758d-3ce5-4fa7-8d96-6b960f029fa9',
            '0780da08-a12b-452a-b936-ebc576aa7670',
            'ff812ca5-ce60-44ac-b07e-66c2c37e98eb'
        ]
        ses = [{'url': f'https://website.org/foo/{x}'} for x in uuids]
        ez = LazyId(ses)
        self.assertEqual(len(uuids), len(ez))
        self.assertCountEqual(map(str, ez), uuids)
        self.assertEqual(ez[0], uuids[0])
        self.assertEqual(ez[0:2], uuids[0:2])
        ez = LazyId([{'id': x} for x in uuids])
        self.assertCountEqual(map(str, ez), uuids)



"""Session creation and datasets registration

The RegistrationClient provides an high-level API for creating experimentation sessions on Alyx
and registering associated datasets.

Summary of methods
------------------
create_new_session - Create a new local session folder and optionally create session record on Alyx
create_session - Create session record on Alyx from local path, without registering files
create_sessions - Create sessions and register files for folder containing a given flag file
register_session - Create a session on Alyx from local path and register any ALF datasets present
register_files - Register a list of files to their respective sessions on Alyx
"""

import pathlib
import uuid
from pathlib import Path, PurePosixPath
import datetime
from logging import getLogger
from uuid import UUID
import itertools
from collections import defaultdict
from fnmatch import fnmatch
import os
from typing import Dict, List
from requests.exceptions import HTTPError
import requests.exceptions
from tqdm import tqdm
from sys import stdout
import pandas as pd
from iblutil.io import hashfile

from .alf.io import next_num_folder
from .alf.files import (
    session_path_parts,
    get_session_path,
    full_path_parts,
    rel_path_parts,
)
from .alf.spec import to_full_path, is_uuid_string
from .alf.exceptions import AlyxSubjectNotFound, ALFError
from .util import ensure_list
from .webclient import no_cache
from .params import get as get_one_params
from . import models


class RegistrationClient:
    """
    Object that keeps the ONE instance and provides method to create sessions and register data.
    """

    def __init__(self, one=None):
        self.one = one
        if not one:
            from .api import ONE

            self.one = ONE(cache_rest=None)
        self.dtypes = self.one.alyx.rest("dataset-types", "list")
        self.registration_patterns = [dt["filename_pattern"] for dt in self.dtypes if dt["filename_pattern"]]
        self.file_extensions = [
            df["file_extension"] for df in self.one.alyx.rest("data-formats", "list", no_cache=True)
        ]

    def create_sessions(self, root_data_folder, glob_pattern="**/create_me.flag", dry=False):
        """
        Create sessions looking recursively for flag files

        Parameters
        ----------
        root_data_folder : str, pathlib.Path
            Folder to look for sessions
        glob_pattern : str
            Register valid sessions that contain this pattern
        dry : bool
            If true returns list of sessions without creating them on Alyx

        Returns
        -------
        list of pathlib.Paths
            Newly created session paths
        list of dicts
            Alyx session records
        """
        logger = getLogger("registration.create_sessions")
        flag_files = list(Path(root_data_folder).glob(glob_pattern))
        records = []
        for flag_file in flag_files:
            if dry:
                records.append(print(flag_file))
                continue
            logger.info("creating session for " + str(flag_file.parent))
            # providing a false flag stops the registration after session creation
            records.append(self.create_session(flag_file.parent))
            flag_file.unlink()
        return [ff.parent for ff in flag_files], records

    def create_session(self, session_path, **kwargs) -> dict:
        """Create a remote session on Alyx from a local session path, without registering files

        Parameters
        ----------
        session_path : str, pathlib.Path
            The path ending with subject/date/number.
        **kwargs
            Optional arguments for RegistrationClient.register_session.

        Returns
        -------
        dict
            Newly created session record.
        """
        return self.register_session(session_path, file_list=False, **kwargs)[0]

    def create_new_session(self, subject, session_root=None, date=None, register=True, **kwargs):
        """Create a new local session folder and optionally create session record on Alyx

        Parameters
        ----------
        subject : str
            The subject name.  Must exist on Alyx.
        session_root : str, pathlib.Path
            The root folder in which to create the subject/date/number folder.  Defaults to ONE
            cache directory.
        date : datetime.datetime, datetime.date, str
            An optional date for the session.  If None the current time is used.
        register : bool
            If true, create session record on Alyx database
        **kwargs
            Optional arguments for RegistrationClient.register_session.

        Returns
        -------
        pathlib.Path
            New local session path
        uuid.UUID
            The experiment UUID if register is True

        Examples
        --------
        Create a local session only

        >>> session_path, _ = RegistrationClient().create_new_session('Ian', register=False)

        Register a session on Alyx in a specific location

        >>> session_path, eid = RegistrationClient().create_new_session('Sy', '/data/lab/Subjects')

        Create a session for a given date

        >>> session_path, eid = RegistrationClient().create_new_session('Ian', date='2020-01-01')
        """
        assert not self.one.offline, "ONE must be in online mode"
        date = self.ensure_ISO8601(date)  # Format, validate
        # Ensure subject exists on Alyx
        self.assert_exists(subject, "subjects")
        session_root = Path(session_root or self.one.alyx.cache_dir) / subject / date[:10]
        session_path = session_root / next_num_folder(session_root)
        session_path.mkdir(exist_ok=True, parents=True)  # Ensure folder exists on disk
        eid = UUID(self.create_session(session_path, **kwargs)["url"][-36:]) if register else None
        return session_path, eid

    def find_files(self, session_path):
        """
        Returns an generator of file names that match one of the dataset type patterns in Alyx

        Parameters
        ----------
        session_path : str, pathlib.Path
            The session path to search

        Returns
        -------
        generator
            Iterable of file paths that match the dataset type patterns in Alyx
        """
        session_path = Path(session_path)
        types = (x["filename_pattern"] for x in self.dtypes if x["filename_pattern"])
        dsets = itertools.chain.from_iterable(session_path.rglob(x) for x in types)
        return (x for x in dsets if x.is_file() and any(x.name.endswith(y) for y in self.file_extensions))

    def assert_exists(self, member, endpoint):
        """Raise an error if a given member doesn't exist on Alyx database

        Parameters
        ----------
        member : str, uuid.UUID, list
            The member ID(s) to verify
        endpoint: str
            The endpoint at which to look it up

        Examples
        --------
        >>> client.assert_exists('ALK_036', 'subjects')
        >>> client.assert_exists('user_45', 'users')
        >>> client.assert_exists('local_server', 'repositories')

        Raises
        -------
        one.alf.exceptions.AlyxSubjectNotFound
            Subject does not exist on Alyx
        one.alf.exceptions.ALFError
            Member does not exist on Alyx
        requests.exceptions.HTTPError
            Failed to connect to Alyx database or endpoint not found
        """
        if isinstance(member, (str, uuid.UUID)):
            try:
                self.one.alyx.rest(endpoint, "read", id=str(member), no_cache=True)
            except requests.exceptions.HTTPError as ex:
                if ex.response.status_code != 404:
                    raise ex
                elif endpoint == "subjects":
                    raise AlyxSubjectNotFound(member)
                else:
                    raise ALFError(f'Member "{member}" doesn\'t exist in Alyx')
        else:
            for x in member:
                self.assert_exists(x, endpoint)

    @staticmethod
    def ensure_ISO8601(date) -> str:
        """Ensure provided date is ISO 8601 compliant

        Parameters
        ----------
        date : str, None, datetime.date, datetime.datetime
            An optional date to convert to ISO string.  If None, the current datetime is used.

        Returns
        -------
        str
            The datetime as an ISO 8601 string
        """
        date = date or datetime.datetime.now()  # If None get current time
        if isinstance(date, str):
            date = datetime.datetime.fromisoformat(date)  # Validate by parsing
        elif type(date) is datetime.date:
            date = datetime.datetime.fromordinal(date.toordinal())
        return datetime.datetime.isoformat(date)

    def register_session(self, ses_path, users=None, file_list=True, **kwargs):
        """
        Register session in Alyx

        NB: If providing a lab or start_time kwarg, they must match the lab (if there is one)
        and date of the session path.

        Parameters
        ----------
        ses_path : str, pathlib.Path
            The local session path
        users : str, list
            The user(s) to attribute to the session
        file_list : bool, list
            An optional list of file paths to register.  If True, all valid files within the
            session folder are registered.  If False, no files are registered
        location : str
            The optional location within the lab where the experiment takes place
        procedures : str, list
            An optional list of procedures, e.g. 'Behavior training/tasks'
        n_correct_trials : int
            The number of correct trials (optional)
        n_trials : int
            The total number of completed trials (optional)
        json : dict, str
            Optional JSON data
        projects: str, list
            The project(s) to which the experiment belongs (optional)
        type : str
            The experiment type, e.g. 'Experiment', 'Base'
        task_protocol : str
            The task protocol (optional)
        lab : str
            The name of the lab where the session took place.  If None the lab name will be
            taken from the path.  If no lab name is found in the path (i.e. no <lab>/Subjects)
            the default lab on Alyx will be used.
        start_time : str, datetime.datetime
            The precise start time of the session.  The date must match the date in the session
            path.
        end_time : str, datetime.datetime
            The precise end time of the session.

        Returns
        -------
        dict
            An Alyx session record
        list, None
            Alyx file records (or None if file_list is False)

        Raises
        ------
        AssertionError
            Subject does not exist on Alyx or provided start_time does not match date in
            session path.
        ValueError
            The provided lab name does not match the one found in the session path or
            start_time/end_time is not a valid ISO date time.
        requests.HTTPError
            A 400 status code means the submitted data was incorrect (e.g. task_protocol was an
            int instead of a str); A 500 status code means there was a server error.
        ConnectionError
            Failed to connect to Alyx, most likely due to a bad internet connection.
        """

        logger = getLogger("registration.register_session")

        if isinstance(ses_path, str):
            ses_path = Path(ses_path)
        details = session_path_parts(ses_path.as_posix(), as_dict=True, assert_valid=True)
        # query alyx endpoints for subject, error if not found
        self.assert_exists(details["subject"], "subjects")

        # look for a session from the same subject, same number on the same day
        with no_cache(self.one.alyx):
            session_id, session = self.one.search(
                subject=details["subject"],
                date_range=details["date"],
                number=details["number"],
                details=True,
                query_type="remote",
            )
        users = ensure_list(users or self.one.alyx.user)
        self.assert_exists(users, "users")

        # if nothing found create a new session in Alyx
        ses_ = {
            "subject": details["subject"],
            "users": users,
            "type": "Experiment",
            "number": details["number"],
        }
        if kwargs.get("end_time", False):
            ses_["end_time"] = self.ensure_ISO8601(kwargs.pop("end_time"))
        start_time = self.ensure_ISO8601(kwargs.pop("start_time", details["date"]))
        assert start_time[:10] == details["date"], "start_time doesn't match session path"
        if kwargs.get("procedures", False):
            ses_["procedures"] = ensure_list(kwargs.pop("procedures"))
        if kwargs.get("projects", False):
            ses_["projects"] = ensure_list(kwargs.pop("projects"))
        assert ("subject", "number") not in kwargs
        if "lab" not in kwargs and details["lab"]:
            kwargs.update({"lab": details["lab"]})
        elif details["lab"] and kwargs.get("lab", details["lab"]) != details["lab"]:
            names = (kwargs["lab"], details["lab"])
            raise ValueError('lab kwarg "%s" does not match lab name in path ("%s")' % names)
        ses_.update(kwargs)

        if not session:  # Create from scratch
            ses_["start_time"] = start_time
            session = self.one.alyx.rest("sessions", "create", data=ses_)
        else:  # Update existing
            if start_time:
                ses_["start_time"] = self.ensure_ISO8601(start_time)
            session = self.one.alyx.rest("sessions", "update", id=session_id[0], data=ses_)

        logger.info(session["url"] + " ")
        # at this point the session has been created. If create only, exit
        if not file_list:
            return session, None
        recs = self.register_files(self.find_files(ses_path) if file_list is True else file_list)
        if recs:  # Update local session data after registering files
            session["data_dataset_session_related"] = ensure_list(recs)
        return session, recs

    def register_water_administration(self, subject, volume, **kwargs):
        """
        Register a water administration to Alyx for a given subject

        Parameters
        ----------
        subject : str
            A subject nickname that exists on Alyx
        volume : float
            The total volume administrated in ml
        date_time : str, datetime.datetime, datetime.date
            The time of administration.  If None, the current time is used.
        water_type : str
            A water type that exists in Alyx; default is 'Water'
        user : str
            The user who administrated the water.  Currently logged-in user is the default.
        session : str, UUID, pathlib.Path, dict
            An optional experiment ID to associate
        adlib : bool
            If true, indicates that the subject was given water ad libitum

        Returns
        -------
        dict
            A water administration record

        Raises
        ------
        one.alf.exceptions.AlyxSubjectNotFound
            Subject does not exist on Alyx
        one.alf.exceptions.ALFError
            User does not exist on Alyx
        ValueError
            date_time is not a valid ISO date time or session ID is not valid
        requests.exceptions.HTTPError
            Failed to connect to database, or submitted data not valid (500)
        """
        # Ensure subject exists
        self.assert_exists(subject, "subjects")
        # Ensure user(s) exist
        user = ensure_list(kwargs.pop("user", [])) or self.one.alyx.user
        self.assert_exists(user, "users")
        # Ensure volume not zero
        if volume == 0:
            raise ValueError("Water volume must be greater than zero")
        # Post water admin
        wa_ = {
            "subject": subject,
            "date_time": self.ensure_ISO8601(kwargs.pop("date_time", None)),
            "water_administered": float(f"{volume:.4g}"),  # Round to 4 s.f.
            "water_type": kwargs.pop("water_type", "Water"),
            "user": user,
            "adlib": kwargs.pop("adlib", False),
        }
        # Ensure session is valid; convert to eid
        if kwargs.get("session", False):
            wa_["session"] = self.one.to_eid(kwargs.pop("session"))
            if not wa_["session"]:
                raise ValueError("Failed to parse session ID")

        return self.one.alyx.rest("water-administrations", "create", data=wa_)

    def register_weight(self, subject, weight, date_time=None, user=None):
        """
        Register a subject weight to Alyx

        Parameters
        ----------
        subject : str
            A subject nickname that exists on Alyx
        weight : float
            The subject weight in grams
        date_time : str, datetime.datetime, datetime.date
            The time of weighing.  If None, the current time is used.
        user : str
            The user who performed the weighing.  Currently logged-in user is the default.

        Returns
        -------
        dict
            An Alyx weight record

        Raises
        ------
        one.alf.exceptions.AlyxSubjectNotFound
            Subject does not exist on Alyx
        one.alf.exceptions.ALFError
            User does not exist on Alyx
        ValueError
            date_time is not a valid ISO date time or weight < 1e-4
        requests.exceptions.HTTPError
            Failed to connect to database, or submitted data not valid (500)
        """
        # Ensure subject exists
        self.assert_exists(subject, "subjects")
        # Ensure user(s) exist
        user = user or self.one.alyx.user
        self.assert_exists(user, "users")
        # Ensure weight not zero
        if weight == 0:
            raise ValueError("Water volume must be greater than 0")

        # Post water admin
        wei_ = {
            "subject": subject,
            "date_time": self.ensure_ISO8601(date_time),
            "weight": float(f"{weight:.4g}"),  # Round to 4 s.f.
            "user": user,
        }
        return self.one.alyx.rest("weighings", "create", data=wei_)

    def files(self, session, file_list, repository_name=None):
        """
        This method checks for the existence of files, groups them by dataset,
        validates their compliance, checks their session match, and adds their records.

        Args:
            session (object): An object that represents the session to which the files belong.
            file_list (list): A list of files to be checked and processed.
            repository_name (str, optional): The name of the repository where the files are stored. Defaults to None.

        Raises:
            It may raise exceptions if the files do not match the provided session. The exact exceptions depend on the
            implementation of 'assert_files_match_session'.
        """
        logger = getLogger("registration.files")

        if len(file_list) == 0:
            return

        files_df = self.group_files_by_dataset(file_list)

        not_compliant_files = files_df[~files_df.alf_compliant]
        if len(not_compliant_files):
            logger.warning(
                "Some files are not alf compliant, and are skipped for registration : "
                f"{not_compliant_files.full_path.to_list()}"
            )
            # we only keep the compliant files for the next steps
            files_df = files_df[files_df.alf_compliant]

        # checks that all files belong to the session supplied
        self.assert_files_match_session(session, files_df)

        for dataset_name, file_group in files_df.groupby("dataset_name"):
            logger.info(f"Dataset {dataset_name}")
            dataset = self._make_dataset(file_group, session, repository_name)
            if dataset is None:  # registration of a new dataset failed
                continue
            self._add_file_records(file_group, session, dataset)

        self.refresh_session_files(session)

    def refresh_session_files(self, session):
        # update the session object to contain info about the new registered data from the remote database

        # name attribute of the session pd.series is the database session id (aka primary key or pk)

        new_session_data = self.one.search(id=session.name, no_cache=True, details=True)

        # we touch the list object that is inside the data_dataset_session_related key of session
        # we cannot change the cell directly as session is a dataframe view.

        # first we clear
        session["data_dataset_session_related"].clear()
        # then we add new data
        session["data_dataset_session_related"].extend(new_session_data["data_dataset_session_related"])

    def group_files_by_dataset(self, files_list: List[str]) -> Dict[str, pd.DataFrame]:
        def make_dataset_name(row):
            if not row.alf_compliant:
                return ""
            collection_name = row.collection + "/" if row.collection else ""
            dataset_name = collection_name + ".".join([row.object, row.attribute, row.extension])
            return dataset_name

        def make_path(row):
            try:
                return to_full_path(**row)
            except Exception as e:  # if we cannot make the alf path from parts, it's not alf compliant
                return ""

        def make_session_alias(row):
            if not row.alf_compliant:
                return ""
            return rf"{row.subject}/{row.date}/{str(row.number).zfill(3)}"

        def make_alf_compliant_flag(row):
            if row.relative_path == "":
                return False
            return True

        results = [
            {"full_path": file, **full_path_parts(file, as_dict=True, assert_valid=False, absolute=True)}
            for file in files_list
        ]
        files_df = pd.DataFrame(results)

        # 1:-1 to remove elements root, and full_path for creating relative path from alf parts
        files_df["relative_path"] = files_df.iloc[:, 2:].apply(make_path, axis="columns")
        files_df["alf_compliant"] = files_df.apply(make_alf_compliant_flag, axis="columns")
        files_df["dataset_name"] = files_df.apply(make_dataset_name, axis="columns")
        files_df["session"] = files_df.apply(make_session_alias, axis="columns")

        return files_df

    def _make_dataset(self, files_df, session, repository_name=None, dry=False):
        logger = getLogger("registration.make_dataset")
        self.assert_single_repository(files_df)

        if repository_name is None:
            repository_name = self.find_session_repo(files_df.iloc[0]["root"])["name"]
        else:
            if repository_name == "default_repo_path":
                repository_name = session.default_data_repository_name
            else:
                # here find_session_repo act mostly a validator for the existance of that repo name
                # (raising if not found, returning the same otherwise)
                repository_name = self.find_session_repo(repository_name)["name"]

        for unique_item in ["collection", "extension", "object", "attribute"]:
            if len(files_df[unique_item].unique()) != 1:
                raise ValueError(
                    f"Registering several files under a same dataset require them being under the same {unique_item}, "
                    f"but these were found : {files_df[unique_item].unique()}"
                )

        dataset_name = files_df.dataset_name.iloc[0]

        object = files_df["object"].unique()[0]
        attribute = files_df["attribute"].unique()[0]
        dataset_type = object + "." + attribute
        collection = files_df["collection"].unique()[0]
        collection.replace("\\", "/")  # in case there is several folders, hence slashes, they should be unix typed
        extension = "." + files_df["extension"].unique()[0]
        session_eid = session.name

        new_dataset = {
            "created_by": get_one_params().ALYX_LOGIN,
            "dataset_type": dataset_type,
            "data_format": extension,
            "collection": collection,
            "session_pk": session_eid,
            "data_repository": repository_name,
        }

        unique_dataset_keys = ["dataset_type", "collection", "data_format"]
        for existing_dataset in session["data_dataset_session_related"]:
            booleans = [existing_dataset[key] == new_dataset[key] for key in unique_dataset_keys]

            # if all keys are matching, returning the existing dataset
            if all(booleans) is True:
                logger.info(
                    f"The dataset {dataset_name} for session {session.alias} "
                    "was already existing. "
                    "Using it to attach files instead of creating a new one."
                )
                return existing_dataset

        if dry:
            new_dataset.update({"id": None, "file_records": []})

        else:
            # if it doesn't exist, create it
            try:
                models.Dataset.assert_valid.dictionnary(new_dataset)
                new_dataset = self.one.alyx.rest("datasets", "create", data=new_dataset)
                logger.info(f"Registered the new dataset : {dataset_name} for session {session.alias}")
            except HTTPError as e:
                logger.info(
                    f"{type(e).__name__} {e} occured during registration of the new dataset : {dataset_name} "
                    f"for session {session.alias}."
                )
                return None

        return new_dataset

    def _add_file_records(self, files_df, session, dataset):
        logger = getLogger("registration.add_file_records")

        existing_files = [os.path.normpath(item["relative_path"]) for item in dataset["file_records"]]

        files_df["file_exists"] = files_df.relative_path.isin(existing_files)

        dataset_name = files_df.dataset_name.iloc[0]

        alread_registered_files = files_df[files_df.file_exists]
        not_yet_registered_files = files_df[~files_df.file_exists]
        if len(alread_registered_files):
            logger.info(
                f"Found {len(alread_registered_files)} "
                f"files already registered for the dataset {dataset_name}. Skipping them"
            )

        if not len(not_yet_registered_files):
            # no file to register, return
            return []

        logger.info(
            f"Starting registration of {len(not_yet_registered_files)} "
            f"new files to the dataset {dataset_name} for the session {session.alias}."
        )

        new_records = []

        for _, file in tqdm(
            not_yet_registered_files.iterrows(),
            total=len(not_yet_registered_files),
            delay=3,
            desc=f"Registering {len(not_yet_registered_files)} files",
            file=stdout,
        ):
            extra = file.extra

            d = {
                "dataset": dataset["id"],
                "extra": extra,
                "exists": True,
            }
            try:
                models.File.assert_valid.dictionnary(d)
                new_file_record = self.one.alyx.rest("files", "create", data=d)
                new_records.append(new_file_record)
            except HTTPError as e:
                logger.error(
                    f"A {type(e).__name__} {e} occured while trying to register file "
                    f"{file.full_path} to dataset {dataset_name}. Skipping"
                )

        return new_records

    def find_session_repo(self, repository_identifier: str):
        if is_uuid_string(repository_identifier):
            try:
                dataset = self.one.alyx.rest(
                    "data-repository",
                    "list",
                    no_cache=True,
                    id="05baa7e4-5eb5-4214-a008-c9e5331004b0",
                )[0]
            except IndexError:
                raise ValueError(f"No dataset id corresponds to the identifier {repository_identifier}")
            return dataset

        try:
            dataset = self.one.alyx.rest(
                "data-repository",
                "list",
                no_cache=True,
                name=repository_identifier,
            )[0]
            return dataset
        except IndexError:
            pass

        try:
            dataset = self.one.alyx.rest(
                "data-repository",
                "list",
                no_cache=True,
                data_path=os.path.normpath(repository_identifier).replace("\\", "/"),  # make the path unix compliant
            )[0]
            return dataset
        except IndexError:
            pass

        raise ValueError(
            f"No existing data repository was found for the identifier {repository_identifier} "
            " That you supplied. Either check their location and move them, or add this root as a new DataRepository"
        )

    def assert_files_match_session(self, session, files_df):
        sessions = files_df.session.unique()
        if len(sessions) != 1:
            raise ValueError(f"More than one session has been found in the file list : {list(sessions)}")
        if session.alias != sessions[0]:
            raise ValueError(
                f"A single session : {sessions[0]} has been found in the file list, "
                f"but no corresponding to the session {session.alias} supplied."
            )

    def assert_single_repository(self, files_df):
        if len(files_df.root.unique()) != 1:
            raise ValueError(
                "More than one data repository (e.g. the root of the files) has been found in the file list"
            )

    def change_session(self, session, **kwargs):
        models.Session.assert_valid.dictionnary(kwargs)

        self.one.alyx.rest("sessions", "partial_update", id=session.name, data=kwargs)

    def change_file(self, session, file_pk, **kwargs):
        file_pk = self.assert_file_in_session(session, file_pk)

        models.File.assert_valid.dictionnary(kwargs)

        self.one.alyx.rest("files", "partial_update", id=file_pk, data=kwargs)
        self.refresh_session_files(session)

    def change_dataset(self, session, dataset_pk, **kwargs):
        dataset_pk = self.assert_dataset_in_session(session, dataset_pk)

        models.Dataset.assert_valid.dictionnary(kwargs)

        self.one.alyx.rest("datasets", "partial_update", id=dataset_pk, data=kwargs)
        self.refresh_session_files(session)

    def delete_file(self, session, file_pk):
        file_pk = self.assert_file_in_session(session, file_pk)
        self.one.alyx.rest("files", "delete", id=file_pk)
        self.refresh_session_files(session)

    def delete_dataset(self, session, dataset_pk):
        dataset_pk = self.assert_dataset_in_session(session, dataset_pk)
        self.one.alyx.rest("datasets", "delete", id=dataset_pk)
        self.refresh_session_files(session)

    def assert_file_in_session(self, session, file_pk):
        if not any(self.one.list_datasets(session, details=True)["file#"] == file_pk):
            raise ValueError(f"The file {file_pk} is not part of the session {session.alias} registered files")
        return file_pk

    def assert_dataset_in_session(self, session, dataset_pk):
        if not any(self.one.list_datasets(session, details=True)["dataset#"] == dataset_pk):
            raise ValueError(f"The dataset {dataset_pk} is not part of the session {session.alias} registered datasets")
        return dataset_pk



# -*- coding: utf-8 -*-
"""
Created on Wed Feb 15 11:05:19 2023

@author: tjostmou
"""

import os, re

from . import spec, files

VGAT_MOUSE_NO = r"Vgat.*?(\d{1,3}).*jRGECO"
"""A pattern to extract the mouse number from old tdms Vgat files"""

TIFF_FRAME_NO = r".*(\d{5})\.tiff?$"
"""A pattern to extract trial number from tiff files"""

TDMS_TRIALS_TABLE_TIME = r"(\d{4}_\d{2}_\d{2}_\d{2}_\d{2})\.tdms$"
"""Extract time from tdms file name"""

TDMS_VIDEO_TRIAL_NO = r".*?(\d+)\.tdms$"
"""Extract trial number from tdms video file name"""

FOV_NO = r"^.*?(?:FOV)?(\d+).*\.(?:(?:PNG)|(?:png))$"
"""Extract field of view number from a png file"""

CONDENSED_DATE = r".*?(20\d{4,6}).*?"
"""Get condensed date"""

SESSION_NO = r".*?20\d{4,6}_(\d).*?"
"""Extract session number from a path next to a date"""

ALYX_PATH_EID = r"\w+(?:\\|\/)\d{4}\-\d{2}\-\d{2}(?:\\|\/)\d{3}"
"""Match a unique session id (human readable) like wm25/2022-08-08/001"""


def extract_extra_from_non_alfpath(filename, re_pattern):
    results = []
    matches = re.finditer(re_pattern, filename, re.MULTILINE|re.IGNORECASE)
    
    for matchnum, match in enumerate(matches,  start = 1):
        for groupx, groupmatch in enumerate(match.groups()):
            results.append(groupmatch)
    return results

def apply_name_change(src_list,new_list):
    assert len(src_list) == len(new_list), "Length of lists must match for name conversion"
    
    for src_file , new_file in zip(src_list,new_list) :
        if new_file == '':
            continue
        os.rename(src_file,new_file)
        

def rename_file(filename, alf_info):
    
    alf_info = files
    
    def _finish_file():
        return spec.to_alf(alf_info['object'],
                                   alf_info['attribute'],
                                   alf_info['extension'],
                                   namespace = alf_info['namespace'],
                                   timescale = alf_info['timescale'],
                                   extra = alf_info['extra'])
    
    def get_imaging_type(collec):
        imaging_types = ("pupil",)
        
        for i_type in imaging_types :
            if i_type in collec :
                return i_type
        return None
    
    alf_info["timescale"] = None
    alf_info["namespace"] = None
    alf_info["revision"] = None
    
    if (alf_info["extension"] == "tif" or alf_info["extension"] == 'tiff') :
        alf_info["object"] = "imaging"
        
        if alf_info["collection"] is not None :
            if "imaging_data" in alf_info["collection"] :
                trial_no = extract_extra_from_non_alfpath(filename, TIFF_FRAME_NO)
                
                if trial_no :
                    alf_info["extra"] = trial_no[0].zfill(5)
                    alf_info["attribute"] = "frames"
                    return _finish_file()
            
    elif alf_info["extension"] == "tdms" :
        
        if alf_info["collection"] is None :
            alf_info["object"] = "trials"
            alf_info["attribute"] = "eventTimeline"
            return _finish_file()
        
        elif "behaviour_imaging" in alf_info["collection"] :
            alf_info["object"] = "behaviour"
            alf_info["attribute"] = "video"
            i_type = get_imaging_type(alf_info["collection"])
            
            if i_type is not None :
                alf_info["extra"] = i_type
                trial_no = extract_extra_from_non_alfpath(filename, TDMS_VIDEO_TRIAL_NO)
                
                if trial_no :
                    alf_info["extra"] = append_extra(trial_no[0].zfill(5) ,alf_info["extra"])
                    return _finish_file()
        
    elif alf_info["extension"] == "PNG" or alf_info["extension"] == 'png' :
        alf_info["extension"] = 'png' #enforce lower caps
        FOV_no = extract_extra_from_non_alfpath(filename, FOV_NO)
        
        if FOV_no :
            alf_info["object"] = "imaging"
            alf_info["attribute"] = "fieldOfView"
            alf_info["extra"] = FOV_no[0].zfill(2)
            return _finish_file()

    return None #if none of the renaming directives returned a results above, we don't rename

    
    